"""
MCMC chains

This module is experimental. Expect API changes.
"""

from abc import abstractmethod
from collections.abc import Callable, Sequence
from typing import Generic, Protocol, TypeVar

import jax
import numpy as np

from liesel.option import Option

from ..docs import usedocs
from .epoch import EpochConfig
from .pytree import concatenate_leaves, slice_leaves
from .types import PyTree

TPyTree = TypeVar("TPyTree", bound=PyTree)


class Chain(Protocol[TPyTree]):
    """
    A ``Chain`` stores multiple chucks of pytrees and concatenates them along a
    time axis.

    The ``Chain`` always assume multiple independent chains that are indexed via
    the first axis. The second dimension represents the time. Consequently, the
    leaves in the pytree must have a dimension of two (i.e., [chain, time,
    ...]).

    A ``Chain`` operates on the assumption that all chunks are pytrees with the
    same structure. However, the time dimension is allowed to vary in size.
    """

    @abstractmethod
    def append(self, chunk: TPyTree) -> None:
        """Appends a chunk to the chain."""
        raise NotImplementedError

    @abstractmethod
    def get(self) -> Option[TPyTree]:
        """
        Returns all chunks combined into one pytree.

        The option is none if no samples are in the chain.
        """
        raise NotImplementedError


class EpochChain(Chain[TPyTree]):
    """
    An ``EpochChain`` is a :class:`.Chain` with an associated :class:`.EpochConfig`.

    The implementation must implement thinning. That is,
    if epoch.thinning > 1 and enabled in contructor, the chain must
    safe only every epoch.thinning element
    """

    @property
    @abstractmethod
    def epoch(self) -> EpochConfig:
        """Returns the associated :class:`.EpochConfig`."""
        raise NotImplementedError


# implementations


@usedocs(Chain)
class ListChain(Generic[TPyTree]):
    """Implements the :class:`.Chain` protocol with a list as storage."""

    def __init__(self):
        self._chunks_list: list[TPyTree] = []

    def append(self, chunk: TPyTree) -> None:
        return self._chunks_list.append(chunk)

    def _concatenate(self) -> None:
        combined = concatenate_leaves(self._chunks_list, 1)
        if combined is not None:
            self._chunks_list = [combined]

    def get(self) -> Option[TPyTree]:
        if len(self._chunks_list) == 0:
            return Option(None)
        else:
            self._concatenate()
            return Option(self._chunks_list[0])


@usedocs(EpochChain)
class ListEpochChain(ListChain[TPyTree]):
    """Implements the :class:`.EpochChain` protocol with a list as storage."""

    def __init__(self, epoch: EpochConfig, apply_thinning: bool = False):
        super().__init__()
        self._epoch = epoch
        self._apply_thinning = apply_thinning
        self._states_counter = 1

    @property
    def epoch(self) -> EpochConfig:
        return self._epoch

    def append(self, chunk: TPyTree) -> None:
        """Applies thinning and appends a chunk to the chain."""
        if self._apply_thinning and self.epoch.thinning > 1:
            th = self._epoch.thinning
            size = jax.tree_util.tree_leaves(chunk)[0].shape[1]
            idx = np.arange(size)[(self._states_counter + np.arange(size)) % th == 0]

            self._states_counter += size

            if len(idx) > 0:
                chunk = slice_leaves(chunk, np.s_[:, idx, ...])
                return super().append(chunk)
        else:
            return super().append(chunk)


class EpochChainManager(Generic[TPyTree]):
    """
    An ``EpochChainManager`` is a container for multiple epoch chains.

    The chains can be concatenated over multiple epochs. Thinning defined in epochs
    can be switched on or of with the constructor flag
    """

    def __init__(self, apply_thinning: bool = False) -> None:
        self._chains: list[ListEpochChain[TPyTree]] = []
        self._apply_thinning = apply_thinning

    @property
    def current_epoch(self) -> EpochConfig:
        """Returns the current epoch."""
        return self._chains[-1].epoch

    def advance_epoch(self, epoch: EpochConfig) -> None:
        """Creates and appends a :class:`.ListEpochChain` for the given ``epoch``."""
        new_chain: ListEpochChain[TPyTree] = ListEpochChain(epoch, self._apply_thinning)
        self._chains.append(new_chain)

    @usedocs(ListEpochChain.append)
    def append(self, chunk: TPyTree) -> None:
        self._chains[-1].append(chunk)

    def get_epochs(self) -> Sequence[EpochConfig]:
        """Returns a list of all epochs."""
        return [c.epoch for c in self._chains]

    def get_specific_chain(self, epoch_number: int) -> ListEpochChain[TPyTree]:
        """Returns the chain for the given epoch number."""
        return self._chains[epoch_number]

    def get_current_chain(self) -> ListEpochChain[TPyTree]:
        """Returns the current chain."""
        return self._chains[-1]

    def get_current_epoch(self) -> EpochConfig:
        """Returns the current poch."""
        return self._chains[-1].epoch

    def combine(self, epoch_numbers: Sequence[int]) -> Option[TPyTree]:
        """
        Combines the given epochs and returns all chunks combined into one pytree.

        The option is none if no samples are in the chain.
        """
        chain: ListChain[TPyTree] = ListChain()
        for num in epoch_numbers:
            epoch_chain = self._chains[num]
            chunk = epoch_chain.get()
            if chunk.is_some():
                chain.append(chunk.unwrap())
        return chain.get()

    def combine_all(self) -> Option[TPyTree]:
        """
        Combines the all epochs and returns all chunks combined into one pytree.

        The option is none if no samples are in the chain.
        """
        chain: ListChain[TPyTree] = ListChain()
        for epoch_chain in self._chains:
            chunk = epoch_chain.get()
            if chunk.is_some():
                chain.append(chunk.unwrap())
        return chain.get()

    def combine_filtered(
        self, predicate: Callable[[EpochConfig], bool]
    ) -> Option[TPyTree]:
        """
        Combines the all epochs for which the predicate evaluates to ``True``.
        Returns all chunks combined into one pytree.

        The option is none if no samples are in the chain.
        """
        chain: ListChain[TPyTree] = ListChain()
        for epoch_chain in self._chains:
            if predicate(epoch_chain.epoch):
                chunk = epoch_chain.get()
                if chunk.is_some():
                    chain.append(chunk.unwrap())
        return chain.get()



"""
Kernel-related info, outcome and mixin classes.
"""

from abc import abstractmethod
from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic

import jax

from ..docs import usedocs
from .epoch import EpochState, EpochType
from .pytree import register_dataclass_as_pytree
from .types import (
    Kernel,
    KeyArray,
    ModelInterface,
    ModelState,
    Position,
    TKernelState,
    TTransitionInfo,
    TTuningInfo,
)


@register_dataclass_as_pytree
@dataclass
class DefaultTransitionInfo:
    """A default template for a transition information object."""

    error_code: int
    """Error code for the transition."""
    acceptance_prob: float
    """Acceptance probability of the transition."""
    position_moved: int
    """Indicates whether the transition resulted in acceptance or not."""

    def minimize(self) -> "DefaultTransitionInfo":
        """Minimizes the transitioninfo."""
        return self


@register_dataclass_as_pytree
@dataclass
class DefaultTuningInfo:
    """A default template for a tuning information object."""

    error_code: int
    """Error code for error during tuning."""
    time: int
    """MCMC time when the tuning happend."""


@register_dataclass_as_pytree
@dataclass
class TransitionOutcome(Generic[TKernelState, TTransitionInfo]):
    """
    A dataclass for the return value of the kernel method :meth:`.Kernel.transition`.
    Different kernels can use different types of :class:`.KernelState`'s and
    :class:`.TransitionInfo`'s.
    """

    info: TTransitionInfo
    """
    A transition info object, see :class:`.DefaultTransitionInfo`.
    """
    kernel_state: TKernelState
    """
    A kernel state object, see :class:`.DAKernelState`.
    """
    model_state: ModelState
    """
    Model state that results from the transition.

    The exact definition depends on the model being used. See, for example,
    :class:`.DictModel`.

    """


@register_dataclass_as_pytree
@dataclass
class TuningOutcome(Generic[TKernelState, TTuningInfo]):
    """
    A dataclass for the return value of the kernel method :meth:`.Kernel.tune`.
    Different kernels can use different types of :class:`.KernelState`'s and
    :class:`.TuningInfo`'s.
    """

    info: TTuningInfo
    """
    A tuning info object, see :class:`.DefaultTuningInfo`.
    """
    kernel_state: TKernelState
    """
    A kernel state object, see :class:`.DAKernelState`.
    """


@register_dataclass_as_pytree
@dataclass
class WarmupOutcome(Generic[TKernelState]):
    """
    A dataclass for the return value of the kernel method :meth:`.Kernel.end_warmup`.
    Different kernels can use different types of :class:`.KernelState`'s.
    """

    error_code: int
    """Error code for the transition."""
    kernel_state: TKernelState
    """
    A kernel state object, see :class:`.DAKernelState`.
    """


class ModelMixin:
    """
    A mixin facilitating the interaction with the model interface.
    """

    position_keys: tuple[str, ...]
    _model: ModelInterface | None

    @property
    def model(self) -> ModelInterface:
        """Returns the model interface if it is set. Raises error otherwise."""

        if self._model is None:
            raise RuntimeError("Model interface not set")

        return self._model

    def set_model(self, model: ModelInterface):
        """Sets the model interface."""
        self._model = model

    def has_model(self) -> bool:
        """Whether the model interface is set."""
        return self._model is not None

    def position(self, model_state: ModelState) -> Position:
        """
        Extracts the position from a model state.
        """

        return self.model.extract_position(self.position_keys, model_state)

    def log_prob_fn(self, model_state: ModelState) -> Callable[[Position], float]:
        """
        Returns the log-probability function with the position as the only argument.
        """

        def log_prob_fn(position: Position) -> float:
            new_model_state = self.model.update_state(position, model_state)
            return self.model.log_prob(new_model_state)

        return log_prob_fn


class TransitionMixin(Generic[TKernelState, TTransitionInfo]):
    """
    An abstract mixin defining two transition methods with and without adaptation.
    """

    @usedocs(Kernel.transition)
    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        is_adaptation = EpochType.is_adaptation(epoch.config.type)

        outcome: TransitionOutcome[TKernelState, TTransitionInfo] = jax.lax.cond(
            is_adaptation,
            self._adaptive_transition,
            self._standard_transition,
            prng_key,
            kernel_state,
            model_state,
            epoch,
        )

        return outcome

    @abstractmethod
    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """
        Performs an MCMC transition *outside* an adaptation epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """
        Performs an MCMC transition *in* an adaptation epoch. Must be jittable.
        """

        raise NotImplementedError


class TuningMixin(Generic[TKernelState, TTuningInfo]):
    """
    An abstract mixin defining two tuning methods after a slow and a fast adaptation
    epoch.
    """

    @usedocs(Kernel.tune)
    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        is_slow = epoch.config.type == EpochType.SLOW_ADAPTATION

        outcome: TuningOutcome[TKernelState, TTuningInfo] = jax.lax.cond(
            is_slow,
            self._tune_slow,
            self._tune_fast,
            prng_key,
            kernel_state,
            model_state,
            epoch,
            history,
        )

        return outcome

    @abstractmethod
    def _tune_fast(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        Tunes a kernel after a *fast* adaptation epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def _tune_slow(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        Tunes a kernel after a *slow* adaptation epoch. Must be jittable.
        """

        raise NotImplementedError



import arviz as az

from .. import __version__
from ..goose.engine import SamplingResults


def to_arviz_inference_data(
    results: SamplingResults, include_warmup: bool = False
) -> az.InferenceData:
    """
    Converts goose's SamplingResults into InferenceData from arviz.

    Arviz' InferenceData seperates samples from the posterior and the warmup in
    the groups 'posterior' and 'warmup_posterior'. By default, all summaries and
    plots use only the data in the group 'posterior'.

    Parameters
    ----------
    results
        The sampling results.
    include_warmup
        Whether to include the warmup in the returned object.


    Returns
    -------
    The inference data.


    Notes
    -----
    The inference data has a variable for each position key included in the
    SamplingResult object. These are usually the position keys of the sampled
    parameters. Goose can track more values if specified in the field
    ``position_included``. This might be helpful to let arviz calculate
    information criteria like WAIC. Assuming that the position key
    ``loglik_pointwise`` corresponds to the point-wise evaluated log-likelihood
    then the inference data can be slightly changed so that arviz interprets
    these values as intended.

    .. code-block:: python

        # re-interpret the data
        llpw_extracted = idat.posterior['loglik_pointwise']
        idat.posterior = idat.posterior.drop('loglik_pointwise')
        idat.add_groups({'log_likelihood': {'observed': llpw_extracted}})
        az.waic(idat)

        # Computed from 300 posterior samples and 10 observations log-likelihood matrix.
        #              Estimate     SE
        # elpd_waic.     -17.43   3.40
        # p_waic           0.16      -

    """

    posterior_samples = results.get_posterior_samples()

    warmup_posterior_samples = None
    if include_warmup:
        warmup_posterior_samples = results.positions.combine_filtered(
            lambda ec: ec.type.is_warmup(ec.type)
        ).expect("No warmup samples found.")

    inference_data = az.from_dict(
        posterior=posterior_samples,
        warmup_posterior=warmup_posterior_samples,
        posterior_attrs={
            "inference_library": "liesel",
            "inference_library_version": __version__,
            "creation_library": "liesel",
            "creation_library_version": __version__,
        },
        save_warmup=include_warmup,
    )

    return inference_data



"""
The degenerate, i.e. rank-deficient, multivariate normal distribution.
"""

from __future__ import annotations

from functools import cached_property
from typing import Any

import jax
import jax.numpy as jnp
import tensorflow_probability.substrates.jax.distributions as tfd
from tensorflow_probability.python.internal import reparameterization
from tensorflow_probability.substrates.jax import tf2jax as tf

Array = Any


def _rank(eigenvalues: Array, tol: float = 1e-6) -> Array | float:
    """
    Computes the rank of a matrix based on the provided eigenvalues. The rank is taken
    to be the number of non-zero eigenvalues.

    Can handle batches.
    """
    mask = eigenvalues > tol
    rank = jnp.sum(mask, axis=-1)
    return rank


def _log_pdet(
    eigenvalues: Array, rank: Array | float | None = None, tol: float = 1e-6
) -> Array | float:
    """
    Computes the log of the pseudo-determinant of a matrix based on the provided
    eigenvalues. If the rank is provided, it is used to select the non-zero eigenvalues.
    If the rank is not provided, it is computed by counting the non-zero eigenvalues. An
    eigenvalue is deemed to be non-zero if it is greater than the numerical tolerance
    ``tol``.

    Can handle batches.
    """
    if rank is None:
        mask = eigenvalues > tol
    else:
        max_index = eigenvalues.shape[-1] - rank

        def fn(i, x):
            return x.at[..., i].set(i >= max_index)

        mask = jax.lax.fori_loop(0, eigenvalues.shape[-1], fn, eigenvalues)

    selected = jnp.where(mask, eigenvalues, 1.0)
    log_pdet = jnp.sum(jnp.log(selected), axis=-1)
    return log_pdet


class MultivariateNormalDegenerate(tfd.Distribution):
    """
    A potentially degenerate multivariate normal distribution.

    Provides the alternative constructor :meth:`.from_penalty` and sampling
    via :meth:`.sample`.

    This is a simplified code-based illustration of how the log-probability for an array
    ``x`` is evaluated::

        xc = x - loc
        log_prob = -0.5 * (rank * np.log(2*np.pi) - log_pdet) -0.5 * (xc.T @ prec @ xc)


    Parameters
    ----------
    loc
        The location (= mean) vector.
    prec
        The precision matrix (= a pseudo-inverse of the variance-covariance matrix).
    rank
        The rank of the precision matrix. Optional.
    log_pdet
        The log-pseudo-determinant of the precision matrix. Optional.
    validate_args
        Python ``bool``, default ``False``. When ``True``, distribution parameters \
        are checked for validity despite possibly degrading runtime performance. \
        When ``False``, invalid inputs may silently render incorrect outputs.
    allow_nan_stats
        Python ``bool``, default ``True``. When ``True``, statistics (e.g., mean, \
        mode, variance) use the value ``NaN`` to indicate the result is undefined. \
        When ``False``, an exception is raised if one or more of the statistic's \
        batch members are undefined.
    name
        Python ``str``, name prefixed to ``Ops`` created by this class.
    tol
        Numerical tolerance for determining which eigenvalues of the distribution's \
        precision matrices should be treated as zeros. Used in :attr:`.rank` and \
        :attr:`.log_pdet`, if they are computed by the class. Also used in \
        :meth:`.sample`.

    Notes
    -----
    * If they are not provided as arguments, ``rank`` and ``log_pdet`` are computed
      based on the eigenvalues of the precision matrix ``prec``. This is an expensive
      operation and can be avoided by specifying the corresponding arguments.
    * When you draw samples from the distribution via :meth:`.sample`, it is always
      necessary to compute the eigendecomposition of the distribution's precision
      matrices once and cache it, because sampling requires both the eigenvalues and
      eigenvectors.

    **Details on sampling**

    To draw samples from a denegerate multivariate normal distribution, we 1) draw
    standard normal samples with mean zero and variance one, 2) transform these samples
    to have the desired covariance structure, and 3) add the desired mean.

    The main problem is to find out how we have to transform the standard normal samples
    in step 2. Say that we have a singular :math:`(m \\times m)`  precision matrix
    :math:`P`. We can view it as
    the generalized inverse of a variance-covariance matrix :math:`\\Sigma`. We can
    obtain :math:`\\Sigma` by finding the eigenvalue decomposition of the precision
    matrix, i.e.

    .. math::
        P = QA^+Q^T,

    where :math:`Q` is the orthogonal matrix of eigenvectors and :math:`A^+` is the
    diagonal matrix of eigenvalues. Note that, if the precision matrix is singular, then
    :math:`\\text{diag}(A^+)` contains zeroes. Now we take the inverse of the non-zero
    entries of :math:`\\text{diag}(A^+)`, while the zero entries remain at zero,
    resulting in a matrix :math:`A`. We can now write

    .. math::
        \\Sigma = Q A Q^T.

    Now we can go through the three steps in detail. We first draw a vector of the
    desired length :math:`z \\sim N(0, I)` from a standard normal distribution.
    :math:`I` is the identity matrix of appropriate dimension. Next, we transform the
    sample by applying :math:`x = Q A^{1/2}z`, such that :math:`\\text{Cov}(x) =
    \\Sigma`:

    .. math::
        \\text{Cov}(x)  & = Q A^{1/2} I (A^{1/2})^T Q^T \\

                        & = Q A Q^T \\

                        & = \\Sigma.

    In the last step, we add the desired mean :math:`\\mu` to :math:`x`.
    Note that the distribution is not a proper distribution on :math:`\\mathbb{R}^m`,
    where :math:`m` refers to the number of columns and rows of :math:`P`.
    Any vector in the null space of :math:`P` can be added to any
    :math:`x \\in \\mathbb{R}^m` without changing the density. The samples generated
    using the procedure described above are orthogonal to the null space of :math:`P`.
    """

    def __init__(
        self,
        loc: Array,
        prec: Array,
        rank: Array | int | None = None,
        log_pdet: Array | float | None = None,
        validate_args: bool = False,
        allow_nan_stats: bool = True,
        name: str = "MultivariateNormalDegenerate",
        tol: float = 1e-6,
    ):
        parameters = dict(locals())

        self._tol = tol
        self._rank = rank
        self._log_pdet = log_pdet

        # necessary for correct broadcasting over event size
        loc = jnp.atleast_1d(loc)

        if not prec.shape[-2] == prec.shape[-1]:
            raise ValueError(
                "`prec` must be square (the last two dimensions must be equal)."
            )

        try:
            jnp.broadcast_shapes(prec.shape[-1], loc.shape[-1])
        except ValueError:
            raise ValueError(
                f"The event sizes of `prec` ({prec.shape[-1]}) and `loc` "
                f"({loc.shape[-1]}) cannot be broadcast together. If you "
                "are trying to use batches for `loc`, you may need to add a "
                "dimension for the event size."
            )

        prec_batches = jnp.shape(prec)[:-2]
        loc_batches = jnp.shape(loc)[:-1]
        self._broadcast_batch_shape = jnp.broadcast_shapes(prec_batches, loc_batches)
        nbatch = len(self.batch_shape)

        self._prec = jnp.expand_dims(prec, tuple(range(nbatch - len(prec_batches))))
        self._loc = jnp.expand_dims(loc, tuple(range(nbatch - len(loc_batches))))

        super().__init__(
            dtype=prec.dtype,
            reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
            parameters=parameters,
            name=name,
        )

    @classmethod
    def from_penalty(
        cls,
        loc: Array,
        var: Array,
        pen: Array,
        rank: Array | int | None = None,
        log_pdet: Array | float | None = None,
        validate_args: bool = False,
        allow_nan_stats: bool = True,
        name: str = "MultivariateNormalDegenerate",
    ) -> MultivariateNormalDegenerate:
        """
        Alternative constructor based on a penalty matrix and an inverse smoothing
        parameter.

        Sometimes, the precision matrix of a degenerate multivariate normal
        distribution is decomposed into a penalty matrix ``pen`` and an inverse
        smoothing parameter ``var``. Using this constructor, a degenerate multivariate
        normal distribution can be initialized from such a decomposition.


        Parameters
        ----------
        loc
            The location (= mean) vector.
        var
            The variance (= inverse smoothing) parameter.
        pen
            The (potentially rank-deficient) penalty matrix.
        rank
            The rank of the penalty matrix. Optional.
        log_pdet
            The log-pseudo-determinant of the penalty matrix. Optional.
        validate_args
            Python ``bool``, default ``False``. When ``True`` distribution parameters
            are checked for validity despite possibly degrading runtime performance.
            When ``False`` invalid inputs may silently render incorrect outputs.
        allow_nan_stats
            Python ``bool``, default ``True``. When ``True``, statistics (e.g., mean,
            mode, variance) use the value ``NaN`` to indicate the result is undefined.
            When ``False``, an exception is raised if one or more of the statistic's
            batch members are undefined.
        name
            Python ``str`` name prefixed to ``Ops`` created by this class.

        Warnings
        --------
        If the log-pseudo-determinant is provided as an argument, it must be
        of the penalty matrix ``pen``, **not** of the precision matrix.

        Notes
        -----
        If they are not provided as arguments, the constructor computes ``rank`` and
        ``log_pdet`` based on the eigenvalues of the penalty matrix ``pen``. This is
        an expensive operation and can be avoided by specifying the corresponding
        arguments.
        """

        prec = pen / jnp.expand_dims(var, axis=(-2, -1))

        evals = jax.numpy.linalg.eigvalsh(pen)
        rank = _rank(evals) if rank is None else rank
        log_pdet = _log_pdet(evals, rank=rank) if log_pdet is None else log_pdet
        log_pdet_prec = log_pdet - rank * jnp.log(var)

        mvnd = cls(
            loc=loc,
            prec=prec,
            rank=rank,
            log_pdet=log_pdet_prec,
            validate_args=validate_args,
            allow_nan_stats=allow_nan_stats,
            name=name,
        )

        return mvnd

    @cached_property
    def eig(self) -> tuple[Array, Array]:
        """Eigenvalues and eigenvectors of the distribution's precision matrices."""
        return jnp.linalg.eigh(self._prec)

    @cached_property
    def _sqrt_pcov(self) -> Array:
        """
        Square roots of the distribution's pseudo-covariance matrices.

        Let ``prec = Q @ A @ Q.T`` be the eigendecomposition of the precision matrix.
        In essence, this property returns ``Q @ jnp.sqrt(1/A)``.
        """
        eigenvalues, evecs = self.eig

        sqrt_eval = jnp.sqrt(1 / eigenvalues)
        sqrt_eval = jnp.where(eigenvalues < self._tol, 0.0, sqrt_eval)

        event_shape = sqrt_eval.shape[-1]
        shape = sqrt_eval.shape + (event_shape,)

        r = tuple(range(event_shape))
        diags = jnp.zeros(shape).at[..., r, r].set(sqrt_eval)
        return evecs @ diags

    @cached_property
    def rank(self) -> Array | float:
        """Ranks of the distribution's precision matrices."""
        if self._rank is not None:
            return self._rank
        evals, _ = self.eig
        return _rank(evals, tol=self._tol)

    @cached_property
    def log_pdet(self) -> Array | float:
        """Log-pseudo-determinants of the distribution's precision matrices."""
        if self._log_pdet is not None:
            return self._log_pdet
        evals, _ = self.eig
        return _log_pdet(evals, self.rank, tol=self._tol)

    @property
    def prec(self) -> Array:
        """Precision matrices."""
        return self._prec

    @property
    def loc(self) -> Array:
        """Locations."""
        return self._loc

    def _sample_n(self, n, seed=None) -> Array:
        shape = [n] + self.batch_shape + self.event_shape

        # The added dimension at the end here makes sure that matrix multiplication
        # with the "sqrt pcov" matrices works out correctly.
        z = jax.random.normal(key=seed, shape=shape + [1])

        # Add a dimension at 0 for the sample size.
        sqrt_pcov = jnp.expand_dims(self._sqrt_pcov, 0)
        centered_samples = jnp.reshape(sqrt_pcov @ z, shape)

        # Add a dimension at 0 for the sample size.
        loc = jnp.expand_dims(self._loc, 0)

        return centered_samples + loc

    def _log_prob(self, x: Array) -> Array | float:
        x = x - self._loc
        # necessary for correct broadcasting in the quadratic form
        x = jnp.expand_dims(x, axis=-2)
        x_T = jnp.swapaxes(x, -2, -1)

        prob1 = -jnp.squeeze(x @ self._prec @ x_T, axis=(-2, -1))
        prob2 = self.rank * jnp.log(2 * jnp.pi) - self.log_pdet
        return 0.5 * (prob1 - prob2)

    def _event_shape(self):
        return tf.TensorShape((jnp.shape(self._prec)[-1],))

    def _event_shape_tensor(self):
        return jnp.array((jnp.shape(self._prec)[-1],), dtype=jnp.int32)

    def _batch_shape(self):
        return tf.TensorShape(self._broadcast_batch_shape)

    def _batch_shape_tensor(self):
        return jnp.array(self._broadcast_batch_shape, dtype=jnp.int32)



"""
MCMC engine

This module is experimental. Expect API changes.
"""

# mypy: check-untyped-defs

from __future__ import annotations

import logging
import pickle
import warnings
from collections.abc import Sequence
from dataclasses import dataclass
from functools import partial
from typing import NamedTuple, cast

import jax
import jax.lax
import jax.numpy as jnp
import jax.random
import jax.tree_util
import numpy as np
from deprecated.sphinx import deprecated
from tqdm import tqdm

from liesel.option import Option

from .chain import Chain, EpochChainManager, ListChain
from .epoch import EpochConfig, EpochManager, EpochState, EpochType
from .kernel_sequence import KernelSequence, KernelStates, TransitionInfos, TuningInfos
from .pytree import as_strong_pytree, register_dataclass_as_pytree
from .types import (
    Array,
    GeneratedQuantity,
    KeyArray,
    ModelInterface,
    ModelState,
    Position,
    PyTree,
    QuantityGenerator,
    TransitionInfo,
)

logger = logging.getLogger(__name__)


class KernelErrorLog(NamedTuple):
    """
    Holds the number of the transitions in which an error in at least one chain occured
    and an array with the error code for each chain.

    Additionally, the kernel identifier is specified and optionally the cls of the
    kernel.
    """

    kernel_ident: str
    kernel_cls: Option[type]  # needed to use the error book

    transition: np.ndarray
    """1-D array (time)."""

    error_codes: np.ndarray
    """2-D array (chain, time)."""


ErrorLog = dict[str, KernelErrorLog]


def _expand_and_stack(chunk, *rest):
    chunks = [chunk]
    chunks.extend(rest)
    expended_chunks = [jnp.expand_dims(chunk, 0) for chunk in chunks]
    return jnp.concatenate(expended_chunks, axis=0)


@deprecated(
    version="0.1.0", reason="Use the functions from liesel.goose.pytree instead."
)
def stack_for_multi(chunks: list):
    """
    Combine identically structured pytrees to be used in multichain.

    The function adds a new dimension (axis 0) to each leaf and stacks the leafs
    along the new axis.
    """

    warnings.warn(
        "``stack_for_multi`` is deprecated. Please use the functions"
        " in the :mod:`.pytree` module.",
        DeprecationWarning,
    )

    return jax.tree_util.tree_map(
        lambda x, *xs: _expand_and_stack(x, *xs), chunks[0], *chunks[1:]
    )


@partial(jax.jit, static_argnums=1)
def _split_keys(keys, n):
    keys = jax.lax.map(lambda key: jax.random.split(key, n), keys)
    return keys


def _initialze_prng(seed: int | KeyArray) -> KeyArray:
    if jnp.isscalar(seed):
        return jax.random.PRNGKey(seed)  # type: ignore
    elif jnp.shape(seed) == (2,):  # type: ignore
        return seed  # type: ignore
    else:
        raise ValueError("Seed has an unsupported shape")


def _add_time_dimension(x: PyTree) -> PyTree:
    """
    Adds a new dimension for time to each leaf.

    The returned tree has the same structure with one additional dimension of
    size 1. The new dimension is ``axis=1``. Each leaf must have at least one
    dimension (representing the chain index).
    """
    initial_position = jax.tree_util.tree_map(
        lambda y, *_ys: jnp.expand_dims(y, 1),
        x,
    )
    return initial_position


@register_dataclass_as_pytree
@dataclass(frozen=True)
class Carry:
    """
    Holds the state that needs to be carried between MCMC interations.
    """

    kernel_states: KernelStates
    model_state: ModelState
    epoch: EpochState


@dataclass
class SamplingResults:
    """
    Contains the results of the MCMC engine.

    Easy access to the samples is provided via the methods
    :func:`.get_samples` and :func:`.get_posterior_samples`.
    """

    positions: EpochChainManager
    """EpochChainManager giving access to monitored variables."""
    transition_infos: EpochChainManager
    """EpochChainManager storing all transition infos."""
    generated_quantities: Option[EpochChainManager]
    """
    Option[EpochChainManager] storing all generated_quantities.

    is_none(), if no quantities have been generated.
    """
    tuning_infos: Option[Chain]
    """
    Option[Chain] storing all tuning infos.

    is_none(), if no tuning was executed
    """

    kernel_states: Option[EpochChainManager]
    """
    Option[EpochChainManager] holds all kernel states.

    is_none(), if monitoring kernel states was not requested.
    """

    full_model_states: Option[EpochChainManager]
    """
    Option[EpochChainManager] holds the full model state of each iteration.

    is_none(), if monitoring was not explicitly requested.
    """

    kernel_classes: Option[dict[str, type]]
    """
    Optional map of kernel identifier to the respective kernel type.
    """

    kernels_by_pos_key: Option[dict[str, str]]
    """
    Optional map of position key to identifier of the for sampling responsible
    kernel.
    """

    def get_samples(self) -> Position:
        """
        Returns a dictionary of all samples for all parameters included in the
        position.
        """
        opt: Option[Position] = self.positions.combine_all()
        return opt.expect(f"No samples in {repr(self)}")

    def get_posterior_samples(self) -> Position:
        """
        Returns a dictionary of posterior samples for all parameters included in the
        position.
        """
        opt = self.positions.combine_filtered(
            lambda config: config.type == EpochType.POSTERIOR
        )
        return opt.expect(f"No posterior samples in {repr(self)}")

    def get_kernels_by_pos_key(self) -> dict[str, str]:
        """
        Returns a dict, identifying the kernel used to sample each position.

        The dict has the format ``{"position name": "kernel identifier"}``.
        """
        return self.kernels_by_pos_key.expect(
            f"No position-kernel associations in {repr(self)}"
        )

    def get_posterior_transition_infos(self) -> dict[str, TransitionInfo]:
        """
        Returns a dictionary of posterior transition information for all parameters
        included in the position.
        """
        opt = self.transition_infos.combine_filtered(
            lambda config: config.type == EpochType.POSTERIOR
        )
        return opt.expect(f"No posterior transition infos in {repr(self)}")

    def get_tuning_times(self) -> Option[Array]:
        """
        Returns array of tuning times.
        """
        if self.tuning_infos.is_none():
            return Option.none()

        # opt_tis is not None since self.tuning_infos is not None
        opt_tis = self.tuning_infos.unwrap().get().unwrap()

        time: Array = next(iter(opt_tis.values())).time

        return Option(time)

    def get_error_log(self, posterior_only=False) -> Option[ErrorLog]:
        """
        Returns the error log that is an dict[kernel_name, KernelErrorLog]
        """
        opt: Option[TransitionInfos]
        if posterior_only:
            opt = self.transition_infos.combine_filtered(
                lambda config: config.type == EpochType.POSTERIOR
            )
            if opt.is_none():
                return Option(None)
            else:
                tis = opt.expect(f"No posterior transition infos in {repr(self)}")
        else:
            opt = self.transition_infos.combine_all()
            tis = opt.expect(f"No transition infos in {repr(self)}")

        error_log: ErrorLog = {}
        for ker_name in tis:
            mask = np.any(tis[ker_name].error_code != 0, axis=0)
            transition: np.ndarray = np.where(mask)[0]
            # cast is ok since the object has more dimensions in the leaf
            error_codes: np.ndarray = cast(np.ndarray, tis[ker_name].error_code)[
                :, mask
            ]
            cls = self.kernel_classes.map(lambda d: d[ker_name])
            error_log[ker_name] = KernelErrorLog(ker_name, cls, transition, error_codes)
        return Option(error_log)

    def pkl_save(self, path) -> None:
        """Save result as a pickled object under :attr:`.path`."""
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def pkl_load(path) -> SamplingResults:
        """Loads the pickled object from :attr:`.path`."""
        with open(path, "rb") as f:
            return pickle.load(f)


@deprecated(reason="Use SamplingResults", version="0.1.4")
class SamplingResult(SamplingResults):
    """Alias of :class:`.SamplingResults` for backwards compatibility."""

    positions: EpochChainManager
    transition_infos: EpochChainManager
    generated_quantities: Option[EpochChainManager]
    tuning_infos: Option[Chain]
    kernel_states: Option[EpochChainManager]
    full_model_states: Option[EpochChainManager]
    kernel_classes: Option[dict[str, type]]
    kernels_by_pos_key: Option[dict[str, str]]


class Engine:
    """MCMC engine capable of combining multiple transition kernels."""

    def __init__(
        self,
        seeds: KeyArray,
        model_states: ModelState,
        kernel_sequence: KernelSequence,
        epoch_configs: Sequence[EpochConfig],
        jitted_sample_duration: int,
        model: ModelInterface,
        position_keys: Sequence[str] | None,
        minimize_transition_infos: bool = False,
        store_kernel_states: bool = False,
        quantity_generators: Sequence[QuantityGenerator] = [],
        show_progress: bool = True,
    ):
        # fill slots that can be filled directly
        self._inital_states = model_states
        self._seeds = seeds
        self._jitted_sample_duration = jitted_sample_duration
        self._minimize_transition_infos = minimize_transition_infos
        self._store_kernel_states = store_kernel_states
        self._model_states = model_states
        self._quantity_generators = quantity_generators
        self._show_progress = show_progress

        self._kernel_sequence = kernel_sequence
        self._epoch_manager = EpochManager(epoch_configs)
        self._warmup_has_ended = False

        if not position_keys:
            position_keys = [
                key
                for ker in self._kernel_sequence._kernels  # FIXME: use of private field
                for key in ker.position_keys
            ]

        self._position_keys = position_keys
        self._model = model

        # feed in history if at least one kernel requires history for tuning
        #
        # FIXME: automatically fetch position keys
        #
        # fetch kernels' position keys and add them automatically to track them
        # in the position chain
        self._history_required_for_tuning = any(
            [ker.needs_history for ker in self._kernel_sequence._kernels]
        )  # FIXME: use of private field

        self._prng_key = seeds

        # setup storage
        self._position_chain: EpochChainManager = EpochChainManager(apply_thinning=True)
        self._transition_info_chain: EpochChainManager = EpochChainManager()
        self._tuning_info_chain: ListChain = ListChain()
        self._kernel_state_chain: EpochChainManager = EpochChainManager()
        self._quantities_chain: EpochChainManager = EpochChainManager(
            apply_thinning=True
        )

        # initialize kernel state
        keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(self._kernel_sequence.init_states)(
            keys, self._model_states
        )

        # current epoch
        self._epoch: EpochState | None = None

        # jit sample function
        self._sample_many_jitted = jax.jit(
            jax.vmap(
                self._sample_many,
                in_axes=(0, None, 0, 0),
                out_axes=(None, 0, 0, 0, 0, 0, 0),
            )
        )

    @property
    def current_epoch(self) -> EpochState:
        """
        Returns the current epoch.

        Raises a :exc:`.RuntimeError` if no epoch is active.
        """
        if self._epoch is None:
            raise RuntimeError("No active epoch")

        return self._epoch

    def sample_all_epochs(self):
        """
        Runs sampling for all remaining epochs.

        Auto-tuning methods are called automatically.
        """
        while self._epoch_manager.has_more():
            self.sample_next_epoch()

    def sample_next_epoch(self):
        """Runs sampling for the next epoch assuming no epoch is active."""
        self._start_epoch()

        # special treatment for the initial values
        if self.current_epoch.config.type == EpochType.INITIAL_VALUES:
            self._handle_inital_values_epoch()
            return

        self._kernel_start_epoch()

        duration = self.current_epoch.config.duration
        epoch_type = self.current_epoch.config.type.name
        jitted = self._jitted_sample_duration

        if self._show_progress:
            logger.info(
                f"Starting epoch: {epoch_type}, {duration} transitions, "
                f"{jitted} jitted together"
            )

        self._sample_for_duration(duration=duration)
        self._end_epoch()

    def append_epoch(self, epoch: EpochConfig):
        """Appends an epoch to the epochs that should be sampled."""
        self._epoch_manager.append(epoch)

    def is_sampling_done(self) -> bool:
        """Returns true if all configured epochs have been sampled."""
        return not self._epoch_manager.has_more()

    def get_results(self) -> SamplingResults:
        """Returns the results of the sampling process."""
        if self._store_kernel_states:
            ksc = self._kernel_state_chain
        else:
            ksc = None

        if self._quantity_generators:
            gqs = self._quantities_chain
        else:
            gqs = None

        kernels = self._kernel_sequence.get_kernels()
        kernels_cls: dict[str, type] = {ker.identifier: type(ker) for ker in kernels}

        kernels_by_position: dict[str, str] = dict()
        for kernel in kernels:
            kernels_by_position.update(
                {key: kernel.identifier for key in kernel.position_keys}
            )

        return SamplingResults(
            positions=self._position_chain,
            transition_infos=self._transition_info_chain,
            generated_quantities=Option(gqs),
            tuning_infos=Option(self._tuning_info_chain),
            kernel_states=Option(ksc),
            full_model_states=Option(None),
            kernel_classes=Option(kernels_cls),
            kernels_by_pos_key=Option(kernels_by_position),
        )

    def _split_prng_key(self, n: int = 1) -> KeyArray:
        keys = _split_keys(self._prng_key, n + 1)
        self._prng_key = keys[:, 0, :]
        return keys[:, 1:, :]

    def _split_prng_key_one(self) -> KeyArray:
        key = self._split_prng_key(1)
        return key[:, 0, :]

    def _generate_quantity(self):
        if not self._quantity_generators:
            return None

        quants = {}

        for qg in self._quantity_generators:
            key = self._split_prng_key_one()
            gen_f = jax.vmap(qg.generate, in_axes=(0, 0, None))
            quant = gen_f(key, self._model_states, self.current_epoch)
            quants[qg.identifier] = quant

        return quants

    def _handle_inital_values_epoch(self):
        assert self.current_epoch.config.type == EpochType.INITIAL_VALUES
        self.current_epoch.advance_time(1)

        initial_position = _add_time_dimension(
            x=jax.vmap(self._model.extract_position, in_axes=(None, 0))(
                self._position_keys, self._model_states
            ),
        )

        self._position_chain.append(initial_position)

        if self._store_kernel_states:
            ks = _add_time_dimension(x=self._kernel_states)
            self._kernel_state_chain.append(ks)

        if self._quantity_generators:
            quants = self._generate_quantity()
            quants = _add_time_dimension(x=quants)
            self._quantities_chain.append(quants)

        self._epoch = None

    def _start_epoch(self):
        """Advances to the next epoch."""
        if self._epoch is not None:
            raise RuntimeError("Epoch is active and not completed")

        self._epoch = self._epoch_manager.next()

        # invoke end_warmup() for the first non-warmup epoch
        if (
            not self._warmup_has_ended
            and self.current_epoch.config.type == EpochType.POSTERIOR
        ):
            self._end_warmup()

        # advance chains to next epoch
        self._position_chain.advance_epoch(self.current_epoch.config)
        self._transition_info_chain.advance_epoch(self.current_epoch.config)
        self._kernel_state_chain.advance_epoch(self.current_epoch.config)
        self._quantities_chain.advance_epoch(self.current_epoch.config)

    def _kernel_start_epoch(self):
        """Inform kernels about new epoch."""
        keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(
            self._kernel_sequence.start_epoch, in_axes=(0, 0, 0, None)
        )(keys, self._kernel_states, self._model_states, self.current_epoch)

    def _end_warmup(self):
        """
        Ends the warmup sequence.

        Calls :func:`.end_warmup` for each kernel. From now on, only epochs of type
        posterior can follow.
        """
        keys = self._split_prng_key_one()
        tuning_infos: TuningInfos | None = self._tuning_info_chain.get().value

        end_warmup_output = jax.vmap(self._kernel_sequence.end_warmup)(
            keys, self._kernel_states, self._model_states, tuning_infos
        )
        self._kernel_states = end_warmup_output.kernel_states

        # add warnings for the user if there are any non-zero error-code
        for kid, ec in end_warmup_output.error_codes.items():
            if jnp.any(ec != 0):
                logger.warning(f"Warmup error code for {kid}: {ec}")

        logger.info("Finished warmup")

    def _end_epoch(self):
        """
        End epoch.

        Informs kernels about the end of the epoch and initializes the tuning
        if required.
        """
        # ensure that an epoch is active
        epoch = self.current_epoch

        # inform kernels about end of epoch
        end_keys = self._split_prng_key_one()
        self._kernel_states = jax.vmap(
            self._kernel_sequence.end_epoch, in_axes=(0, 0, 0, None)
        )(end_keys, self._kernel_states, self._model_states, epoch)

        self._tune_kernels(epoch)

        if self._show_progress:
            ti_option = self._transition_info_chain.get_current_chain().get()

            def count_non_zero_error_codes(tis: TransitionInfos):
                cts = {}
                for kernel_id, ti in tis.items():
                    nzero = jnp.sum(ti.error_code != 0, axis=1)
                    ntrans = ti.error_code.shape[1]  # type: ignore
                    cts[kernel_id] = (nzero, ntrans)
                return cts

            error_info: dict[str, tuple[Array, int]] = ti_option.map_or(
                {}, count_non_zero_error_codes
            )

            for kid, kcts in error_info.items():
                if jnp.any(kcts[0] != 0):
                    logger.warning(
                        f"Errors per chain for {kid}: "
                        f"{', '.join(map(str, kcts[0]))} / {kcts[1]} transitions"
                    )

            logger.info("Finished epoch")

        # no epoch is active anymore
        self._epoch = None

    def _tune_kernels(self, epoch: EpochState):
        """Trigger tuning if epoch is an adaptation phase."""
        if EpochType.is_adaptation(epoch.config.type):
            tune_keys = self._split_prng_key_one()
            if self._history_required_for_tuning:
                history = (
                    self._position_chain.get_current_chain()
                    .get()
                    .expect("The history must contain samples.")
                )
            else:
                history = None

            tune_output = jax.vmap(
                self._kernel_sequence.tune, in_axes=(0, 0, 0, None, 0)
            )(tune_keys, self._kernel_states, self._model_states, epoch, history)
            self._kernel_states = tune_output.kernel_states

            # we need to add the time dimension
            self._tuning_info_chain.append(_add_time_dimension(x=tune_output.infos))

    def _sample_many(
        self,
        keys: KeyArray,
        epoch: EpochState,
        kernel_states: KernelStates,
        model_state: ModelState,
    ) -> tuple[
        EpochState,
        KernelStates,
        ModelState,
        Position,
        TransitionInfos,
        None | KernelStates,
        None | dict[str, GeneratedQuantity],
    ]:
        def scan_f(
            carry: Carry, key: KeyArray
        ) -> tuple[
            Carry,
            tuple[
                Position,
                TransitionInfos,
                None | KernelStates,
                None | dict[str, GeneratedQuantity],
            ],
        ]:
            key_trans, key_quants = jax.random.split(key)
            epoch = carry.epoch
            out = self._kernel_sequence.transition(
                key_trans, carry.kernel_states, carry.model_state, epoch
            )
            epoch.advance_time(1)
            new_carry = Carry(out.kernel_states, out.model_state, epoch)

            # extract the position specified to store in chain
            position = self._model.extract_position(
                self._position_keys, out.model_state
            )

            # minimize transition infos if requested
            tinfos = out.infos
            if self._minimize_transition_infos:
                for id in tinfos.keys():
                    tinfos[id] = tinfos[id].minimize()

            ks = None
            if self._store_kernel_states:
                ks = new_carry.kernel_states

            quants = None
            if self._quantity_generators:
                quants = {}
                keys = jax.random.split(key_quants, len(self._quantity_generators))
                for i, qg in enumerate(self._quantity_generators):
                    key = keys[i]
                    quant = qg.generate(key, out.model_state, epoch)
                    quants[qg.identifier] = quant

            return new_carry, (position, tinfos, ks, quants)

        inital_carry = Carry(kernel_states, model_state, epoch)
        carry, chain = jax.lax.scan(scan_f, inital_carry, keys)
        kernel_states = carry.kernel_states
        model_state = carry.model_state
        epoch = carry.epoch

        return (
            epoch,
            kernel_states,
            model_state,
            chain[0],
            chain[1],
            chain[2],
            chain[3],
        )

    def _sample_for_duration(self, duration: int):
        if self.current_epoch.time_left() < duration:
            raise RuntimeError("Not enough time left in epoch")

        if duration % self._jitted_sample_duration:
            raise RuntimeError(
                f"Duration {duration} is not a multiple of the "
                f"jitted sampling duration {self._jitted_sample_duration}"
            )

        # convert to non-weak device arrays to avoid recompilation
        self._epoch = as_strong_pytree(self._epoch)
        self._kernel_states = as_strong_pytree(self._kernel_states)
        self._model_states = as_strong_pytree(self._model_states)

        it = range(duration // self._jitted_sample_duration)

        if self._show_progress:
            it = tqdm(it, ncols=80, disable=None, unit="chunk")

        for dur_i in it:
            # FIXME: split for entire duration instead of each loop iteration
            keys = self._split_prng_key(self._jitted_sample_duration)
            (
                new_epoch,
                new_ks,
                new_ms,
                position_chain,
                infos,
                ksc,
                quants,
            ) = self._sample_many_jitted(
                keys, self.current_epoch, self._kernel_states, self._model_states
            )

            self._epoch = new_epoch
            self._kernel_states = new_ks
            self._model_states = new_ms
            self._position_chain.append(position_chain)
            self._transition_info_chain.append(infos)
            if self._store_kernel_states:
                self._kernel_state_chain.append(ksc)
            if self._quantity_generators:
                self._quantities_chain.append(quants)



import pytest

from liesel.option import Option


def test_none() -> None:
    opt: Option[int] = Option(None)
    assert opt.is_none()
    assert not opt.is_some()

    opt2: Option[int] = Option.none()
    assert opt2.is_none()
    assert not opt2.is_some()


def test_some() -> None:
    opt: Option[int] = Option(1)
    assert not opt.is_none()
    assert opt.is_some()

    opt2: Option[int] = Option.some(1)
    assert not opt2.is_none()
    assert opt2.is_some()


def test_unwrap() -> None:
    opt: Option[int] = Option(1)
    assert opt.unwrap() == 1

    opt2: Option[int] = Option.some(1)
    assert opt2.unwrap() == 1

    opt3: Option[int] = Option(None)
    with pytest.raises(RuntimeError):
        opt3.unwrap()


def test_expect() -> None:
    opt: Option[int] = Option(1)
    assert opt.expect("my msg") == 1

    opt2: Option[int] = Option.some(1)
    assert opt2.expect("my msg") == 1

    opt3: Option[int] = Option(None)
    with pytest.raises(RuntimeError):
        opt3.expect("This should fail!")


def test_unwrap_or() -> None:
    opt: Option[int] = Option(1)
    assert opt.unwrap_or(2) == 1

    opt2: Option[int] = Option(None)
    assert opt2.unwrap_or(2) == 2


def test_map() -> None:
    opt: Option[int] = Option(1)
    assert opt.map(lambda x: 2 * x) == Option(2)

    opt2: Option[int] = Option(None)
    assert opt2.map(lambda x: 2 * x) == Option(None)


def test_map_or() -> None:
    opt: Option[int] = Option(1)
    assert opt.map_or(4, lambda x: 2 * x) == 2

    opt2: Option[int] = Option(None)
    assert opt2.map_or(4, lambda x: 2 * x) == 4



import logging
from collections.abc import Generator
from contextlib import contextmanager

import jax.numpy as jnp
import pytest
from _pytest.logging import LogCaptureHandler

from liesel.goose.builder import EngineBuilder
from liesel.goose.engine import SamplingResults
from liesel.goose.epoch import EpochConfig, EpochType
from liesel.goose.models import DictModel

from .mock_kernel import MockKernel


def pytest_addoption(parser):
    parser.addoption(
        "--run-mcmc", action="store_true", default=False, help="run mcmc tests"
    )

    parser.addoption(
        "--mcmc-seed", action="store", default=42, help="set mcmc seed", type=int
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "mcmc: mark test as mcmc test")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-mcmc"):
        # --run-mcmc given in cli: do not skip mcmc tests
        return

    skip_mcmc = pytest.mark.skip(reason="need --run-mcmc option to run")

    for item in items:
        if "mcmc" in item.keywords:
            item.add_marker(skip_mcmc)


@pytest.fixture
def mcmc_seed(request):
    return request.config.getoption("--mcmc-seed")


@contextmanager
def local_caplog_fn(
    level: int = logging.INFO, name: str = "liesel"
) -> Generator[LogCaptureHandler, None, None]:
    """
    Context manager that captures records from non-propagating loggers.

    After the end of the ``with`` statement, the log level is restored to its original
    value. Code adapted from `this GitHub comment <GH_>`_.

    .. _GH: https://github.com/pytest-dev/pytest/issues/3697#issuecomment-790925527

    Parameters
    ----------
    level
        The log level.
    name
        The name of the logger to update.
    """

    logger = logging.getLogger(name)

    old_level = logger.level
    logger.setLevel(level)

    handler = LogCaptureHandler()
    logger.addHandler(handler)

    try:
        yield handler
    finally:
        logger.setLevel(old_level)
        logger.removeHandler(handler)


@pytest.fixture
def local_caplog():
    """
    Fixture that yields a context manager for capturing records from non-propagating
    loggers.

    Examples
    --------
    Usage example::

        import liesel.model.distreg as dr

        def test_build_empty(local_caplog):
            with local_caplog() as caplog:
                drb = dr.DistRegBuilder()
                model = drb.build()
                assert len(caplog.records) == 1
                assert caplog.records[0].levelname == "WARNING"
    """

    yield local_caplog_fn


@pytest.fixture(scope="module")
def result() -> SamplingResults:
    builder = EngineBuilder(0, 3)
    state = {
        "foo": jnp.arange(3, dtype=jnp.float32),
        "bar": jnp.zeros((3, 5, 7)),
        "baz": jnp.array(1.0),
    }

    builder.add_kernel(MockKernel(list(state.keys())))
    builder.set_model(DictModel(log_prob_fn=lambda state: 0.0))
    builder.set_initial_values(state)
    builder.set_epochs(
        [
            EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
            EpochConfig(EpochType.BURNIN, 50, 1, None),
            EpochConfig(EpochType.POSTERIOR, 250, 1, None),
        ]
    )
    engine = builder.build()
    engine.sample_all_epochs()
    return engine.get_results()


@pytest.fixture(scope="module")
def result_for_quants() -> SamplingResults:
    builder = EngineBuilder(0, 4)
    state = {
        "foo": jnp.arange(5, dtype=jnp.float32),
        "bar": jnp.zeros((3, 5, 7)),
        "baz": jnp.array(1.0),
    }

    builder.add_kernel(MockKernel(list(state.keys())))
    builder.set_model(DictModel(log_prob_fn=lambda state: 0.0))
    builder.set_initial_values(state)
    builder.set_epochs(
        [
            EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
            EpochConfig(EpochType.BURNIN, 50, 1, None),
            EpochConfig(EpochType.POSTERIOR, 250, 1, None),
        ]
    )
    engine = builder.build()
    engine.sample_all_epochs()
    return engine.get_results()


@pytest.fixture(scope="module")
def single_chain_result() -> SamplingResults:
    builder = EngineBuilder(0, 1)
    state = {
        "foo": jnp.arange(3, dtype=jnp.float32),
        "bar": jnp.zeros((3, 5, 7)),
        "baz": jnp.array(1.0),
    }

    builder.add_kernel(MockKernel(list(state.keys())))
    builder.set_model(DictModel(log_prob_fn=lambda state: 0.0))
    builder.set_initial_values(state)
    builder.set_epochs(
        [
            EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
            EpochConfig(EpochType.BURNIN, 50, 1, None),
            EpochConfig(EpochType.POSTERIOR, 250, 1, None),
        ]
    )
    engine = builder.build()
    engine.sample_all_epochs()
    return engine.get_results()



"""
Goose MCMC framework.
"""

from .builder import EngineBuilder as EngineBuilder
from .engine import Engine as Engine
from .epoch import EpochConfig, EpochType
from .gibbs import GibbsKernel
from .hmc import HMCKernel
from .iwls import IWLSKernel
from .mh_kernel import MHKernel, MHProposal
from .models import DictModel
from .nuts import NUTSKernel
from .rw import RWKernel
from .summary_m import Summary
from .summary_viz import (
    plot_cor,
    plot_density,
    plot_pairs,
    plot_param,
    plot_scatter,
    plot_trace,
)
from .types import ModelInterface, Position
from .warmup import stan_epochs

__all__ = [
    "DictModel",
    "Engine",
    "EngineBuilder",
    "EpochConfig",
    "EpochType",
    "GibbsKernel",
    "HMCKernel",
    "IWLSKernel",
    "MHKernel",
    "MHProposal",
    "ModelInterface",
    "NUTSKernel",
    "Position",
    "RWKernel",
    "Summary",
    "plot_cor",
    "plot_density",
    "plot_param",
    "plot_trace",
    "plot_scatter",
    "plot_pairs",
    "stan_epochs",
]



from dataclasses import dataclass

import jax
import pytest

from liesel.goose.pytree import register_dataclass_as_pytree


@dataclass
class NoPytree:
    foo: float
    bar: int


@register_dataclass_as_pytree
@dataclass
class Pytree:
    foo: float
    bar: int


@jax.jit
def set(foobar):
    foobar.bar = 2
    foobar.foo = 0.0
    return foobar


def test_dataclass_not_working():
    no_pytree = NoPytree(foo=1.0, bar=0)
    with pytest.raises(TypeError):
        _ = set(no_pytree)


def test_registered_dataclass_is_working():
    pytree = Pytree(foo=1.0, bar=0)

    pytree = set(pytree)

    assert pytree.foo == 0.0
    assert pytree.bar == 2



from typing import Any

import jax.numpy as jnp

from liesel.goose.chain import (
    Chain,
    EpochChain,
    EpochChainManager,
    ListChain,
    ListEpochChain,
)
from liesel.goose.epoch import EpochConfig, EpochType
from liesel.goose.types import PyTree


def create_epoch(dur: int) -> EpochConfig:
    return EpochConfig(
        type=EpochType.POSTERIOR, duration=dur, thinning=1, optional=None
    )


def create_chunk() -> PyTree:
    chunk = (
        jnp.expand_dims(jnp.array([1.0, 2.0, 3.0]), (0, 1)),
        {
            "foo": jnp.expand_dims(jnp.array(1.0), (0, 1)),
            "bar": jnp.expand_dims(jnp.array([[1.0, 2], [3.0, 4.0]]), (0, 1)),
        },
    )
    return chunk


def test_list_chain() -> None:
    chain: Chain = ListChain()
    assert chain.get().is_none()

    chain.append(create_chunk())
    assert chain.get().is_some()

    chain.append(create_chunk())
    assert chain.get().is_some()

    chain.append(create_chunk())
    chain.append(create_chunk())
    chain.append(create_chunk())
    comp = chain.get().unwrap()
    assert comp[0].shape == (1, 5, 3)
    assert comp[1]["foo"].shape == (1, 5)
    assert comp[1]["bar"].shape == (1, 5, 2, 2)


def test_epoch_list_chain():
    chain: EpochChain[Any] = ListEpochChain(create_epoch(11))
    assert chain.get().is_none()

    chain.append(create_chunk())
    assert chain.get().is_some()

    chain.append(create_chunk())
    assert chain.get().is_some()

    chain.append(create_chunk())
    chain.append(create_chunk())
    chain.append(create_chunk())
    comp = chain.get().unwrap()
    assert comp[0].shape == (1, 5, 3)
    assert comp[1]["foo"].shape == (1, 5)
    assert comp[1]["bar"].shape == (1, 5, 2, 2)


def test_epoch_chain_manager() -> None:
    manager: EpochChainManager[Any] = EpochChainManager()

    nepochs = 3
    for i in range(nepochs):
        dur = (1 + i) * 5
        manager.advance_epoch(create_epoch(dur))
        for _ in range(manager.get_current_epoch().duration):
            manager.append(create_chunk())

    # test combine_all
    chain = manager.combine_all().unwrap()
    assert chain[0].shape == (1, 5 + 10 + 15, 3)

    # test combine
    chain = manager.combine((1, 2)).unwrap()
    assert chain[0].shape == (1, 10 + 15, 3)

    # test combine_filtered
    chain = manager.combine_filtered(lambda config: config.duration == 10).unwrap()
    assert chain[0].shape == (1, 10, 3)


def test_epoch_chain_thinning() -> None:
    # thinning larger than chunk size
    chain: ListEpochChain[Any] = ListEpochChain(
        EpochConfig(EpochType.POSTERIOR, 100, 10, None), apply_thinning=True
    )

    data = jnp.arange(0, 10).reshape(2, 5)

    for _ in range(100 // 5):
        chain.append(data)
    res = chain.get().unwrap()

    assert res.shape == (2, 10)
    assert jnp.all(res.sum(axis=1) == 10 * jnp.array((4, 9)))

    # thinning smaller than chunk size
    chain = ListEpochChain(
        EpochConfig(EpochType.POSTERIOR, 99, 3, None), apply_thinning=True
    )

    data = jnp.arange(0, 18).reshape(2, 9)

    for _ in range(99 // 9):
        chain.append(data)
    res = chain.get().unwrap()

    assert res.shape == (2, 33)
    assert jnp.all(
        res.sum(axis=1) == 11 * jnp.array(((2, 5, 8), (11, 14, 17))).sum(axis=1)
    )

    # thinning enabled but set to 1
    chain = ListEpochChain(
        EpochConfig(EpochType.POSTERIOR, 10, 1, None), apply_thinning=True
    )

    data = jnp.arange(0, 10).reshape(2, 5)

    for _ in range(10 // 5):
        chain.append(data)
    res = chain.get().unwrap()

    assert res.shape == (2, 10)
    assert jnp.all(res.sum(axis=1) == 2 * data.sum(axis=1))



"""
Model visualization.
"""

import matplotlib.pyplot as plt
import networkx as nx
from matplotlib.lines import Line2D


def plot_nodes(model, show=True, save_path=None, width=14, height=10, prog="dot"):
    """
    Plots the nodes of a Liesel model.

    Parameters
    ----------
    model
        The model to be plotted.
    show
        Whether to show the plot in a new window.
    save_path
        Path to save the plot. If not provided, the plot will not be saved.
    width
        Width of the plot in inches.
    height
        Height of the plot in inches.
    prog
        Layout parameter. Available layouts: circo, dot (the default), fdp, neato,
        osage, patchwork, sfdp, twopi.
    """

    colors = [
        "#fc8d62" if node.outdated else "#8da0cb" for node in model.node_graph.nodes
    ]

    _, axis, pos = _prepare_figure(model.node_graph, width, height, prog)
    nx.draw_networkx_nodes(model.node_graph, pos, node_color=colors, ax=axis)
    _add_labels(model.node_graph, axis, pos)
    _draw_edges(model.node_graph, axis, pos)

    if save_path:
        plt.savefig(save_path)
    if show:
        plt.show()


def plot_vars(model, show=True, save_path=None, width=14, height=10, prog="dot"):
    """
    Plots the variables of a Liesel model.

    Parameters
    ----------
    model
        The model to be plotted.
    show
        Whether to show the plot in a new window.
    save_path
        Path to save the plot. If not provided, the plot will not be saved.
    width
        Width of the plot in inches.
    height
        Height of the plot in inches.
    prog
        Layout parameter. Available layouts: circo, dot (the default), fdp, neato,
        osage, patchwork, sfdp, twopi.
    """

    _, axis, pos = _prepare_figure(model.var_graph, width, height, prog)
    _add_nodes_with_distribution_to_plot(model.var_graph, axis, pos)
    _add_nodes_without_distribution_to_plot(model.var_graph, axis, pos)
    _add_labels(model.var_graph, axis, pos)
    _draw_edges(model.var_graph, axis, pos)
    _add_legend(axis)

    if save_path:
        plt.savefig(save_path)
    if show:
        plt.show()


def _prepare_figure(graph, width, height, prog):
    """Prepares the figure for plotting."""

    fig, axis = plt.subplots()
    fig.set_size_inches(width, height)

    if _is_pygraphviz_installed():
        pos = nx.drawing.nx_agraph.graphviz_layout(graph, prog=prog)
    else:
        pos = nx.fruchterman_reingold_layout(graph)

    return fig, axis, pos


def _is_pygraphviz_installed():
    """Checks if pygraphviz is installed."""

    try:
        import pygraphviz  # noqa: F401
    except ImportError:
        return False
    else:
        return True


def _add_nodes_with_distribution_to_plot(graph, axis, pos):
    """Adds nodes with distribution to the figure."""

    nodes_with_distribution = {
        node: "#fc8d62" if node.weak else "#8da0cb"
        for node in graph.nodes
        if node.has_dist
    }

    nx.draw_networkx_nodes(
        graph,
        pos,
        node_size=1200,
        node_color=nodes_with_distribution.values(),
        nodelist=nodes_with_distribution,
        node_shape="*",
        ax=axis,
    )


def _add_nodes_without_distribution_to_plot(graph, axis, pos):
    """Adds nodes without distribution to the figure."""

    nodes_without_distribution = {
        node: "#fc8d62" if node.weak else "#8da0cb"
        for node in graph.nodes
        if not node.has_dist
    }

    nx.draw_networkx_nodes(
        graph,
        pos,
        node_size=500,
        node_color=nodes_without_distribution.values(),
        nodelist=nodes_without_distribution,
        node_shape="o",
        ax=axis,
    )


def _add_labels(graph, axis, pos):
    """Adds labels to the figure."""

    labels = {
        node: f"{type(node).__name__}\n{node.name}"
        if node.name is not None
        else node.role.name
        for node in pos
    }

    nx.draw_networkx_labels(graph, pos, labels=labels, ax=axis, font_size=10)


def _draw_edges(graph, axis, pos):
    """Adds edges to the figure."""

    nx.draw_networkx_edges(
        graph,
        pos,
        edgelist=graph.edges,
        edge_color="#aaaaaa",
        arrows=True,
        ax=axis,
        node_size=500,
    )


def _add_legend(axis):
    """Adds a legend to the figure."""

    legend_elements = [
        Line2D([0], [0], color="#8da0cb", lw=4, label="Strong"),
        Line2D([0], [0], color="#fc8d62", lw=4, label="Weak"),
        Line2D(
            [0],
            [0],
            marker="*",
            color="w",
            label="With distribution",
            markerfacecolor="k",
            markersize=18,
        ),
        Line2D(
            [0],
            [0],
            marker="o",
            color="w",
            label="Without distribution",
            markerfacecolor="k",
            markersize=12,
        ),
    ]

    axis.legend(handles=legend_elements, loc="best")



import os.path

import matplotlib
import pytest

from liesel.goose.engine import SamplingResults
from liesel.goose.summary_viz import (
    _setup_plot_df,
    plot_cor,
    plot_density,
    plot_param,
    plot_trace,
)

# use non-interactive backend
matplotlib.use("template")

# load results from file
# file was generated with files/files_for_test_summary.py
path_module_dir = os.path.dirname(__file__)
path = os.path.join(path_module_dir, "files", "summary_viz_res.pkl")
results = SamplingResults.pkl_load(path)


def test_data_complete():
    data_complete = _setup_plot_df(
        results, params=None, param_indices=None, chain_indices=None, max_chains=None
    )

    assert len(data_complete) == 90000
    assert len(data_complete.columns) == 6
    assert data_complete["param"].nunique() == 2
    assert data_complete["param_label"].nunique() == 6

    assert data_complete["param_label"].unique().tolist() == [
        "beta[0]",
        "beta[1]",
        "beta[2]",
        "beta[3]",
        "beta[4]",
        "log_sigma",
    ]

    assert data_complete["chain_index"].nunique() == 15
    assert data_complete["iteration"].nunique() == 1000


def test_data_subset():
    data_subset = _setup_plot_df(
        results,
        params="beta",
        param_indices=[1, 3, 4],
        chain_indices=[0, 1, 10],
        max_chains=None,
    )

    assert len(data_subset) == 9000
    assert len(data_subset.columns) == 6
    assert data_subset["param"].nunique() == 1

    assert data_subset["param_label"].unique().tolist() == [
        "beta[1]",
        "beta[3]",
        "beta[4]",
    ]

    assert data_subset["chain_index"].nunique() == 3
    assert data_subset["iteration"].nunique() == 1000


def test_max_chains():
    data_max_chains_large = _setup_plot_df(
        results, params=None, param_indices=None, chain_indices=None, max_chains=100
    )

    assert data_max_chains_large["chain_index"].nunique() == 15

    data_max_chains_small = _setup_plot_df(
        results,
        params=None,
        param_indices=None,
        chain_indices=[0, 3, 5, 6, 7, 8, 10],
        max_chains=2,
    )

    assert data_max_chains_small["chain_index"].nunique() == 2
    assert data_max_chains_small["chain_index"].unique().tolist() == [0, 3]


def test_plot_defaults():
    plot_trace(results)
    plot_density(results)
    plot_cor(results)
    plot_param(results, param="beta", param_index=3)
    plot_param(results, param="log_sigma")


def test_plot_aesthetics():
    plot_trace(
        results,
        max_chains=3,
        title="Custom Title",
        title_spacing=0.86,
        xlabel="Custom xlabel",
        style="white",
        color_palette=["blue", "red", "green"],
        ncol=4,
        height=4,
        aspect_ratio=1.2,
    )

    plot_density(
        results,
        max_chains=3,
        title="Custom Title",
        title_spacing=0.86,
        xlabel="Custom xlabel",
        style="white",
        color_palette={1: "blue", 0: "green", 2: "red"},
        ncol=3,
        height=4,
        aspect_ratio=1.2,
    )

    plot_cor(
        results,
        max_chains=10,
        max_lags=10,
        title="Custom Title",
        title_spacing=0.86,
        xlabel="Custom xlabel",
        style="white",
        ncol=3,
        height=3,
        aspect_ratio=0.8,
    )

    plot_param(
        results,
        param="beta",
        param_index=1,
        max_chains=3,
        max_lags=10,
        title_spacing=0.91,
        style="white",
        color_list=["blue", "red", "green"],
        figure_size=(15, 10),
        legend_position=(1.1, 0.2),
    )


def test_correct_error_message():
    with pytest.raises(ValueError) as exc_info:
        plot_param(results, param="beta")

    assert exc_info.type is ValueError

    assert (
        exc_info.value.args[0]
        == "beta has more than one index. Please specify a single `param_index` "
        "for plotting."
    )



"""
Type aliases, type variables and protocols.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, ClassVar, NewType, Protocol, TypeVar

if TYPE_CHECKING:
    from .epoch import EpochState
    from .kernel import (
        DefaultTransitionInfo,
        TransitionOutcome,
        TuningOutcome,
        WarmupOutcome,
    )

# simple type aliases

PyTree = Any
Array = Any

ModelState = PyTree
Position = NewType("Position", dict[str, Array])
KernelState = PyTree
KeyArray = Any


class TuningInfo(Protocol):
    """Holds information about sampler tuning."""

    error_code: int
    time: int


class TransitionInfo(Protocol):
    """Holds information about MCMC transitions."""

    error_code: int
    """
    An error code defined in the error book of the kernel.
    0 if no errors occurred during the transition.
    """

    acceptance_prob: float
    """
    The acceptance probability of a proposal of a Metropolis-Hastings kernel
    or the average acceptance probability across a trajectory of a NUTS-type kernel.
    99.0 if not used by the kernel.
    """

    position_moved: int
    """
    0 if the position did not move during the transition, 1 if it *did* move,
    and 99 if unknown.
    """

    @abstractmethod
    def minimize(self) -> DefaultTransitionInfo:
        raise NotImplementedError


TKernelState = TypeVar("TKernelState", bound=KernelState)
TTransitionInfo = TypeVar("TTransitionInfo", bound=TransitionInfo)
TTuningInfo = TypeVar("TTuningInfo", bound=TuningInfo)


class ModelInterface(Protocol):
    """
    Defines a standardized way for Goose to communicate with a statistical model.

    This means predominantly, to update the model state and to compute the
    log-probability.
    """

    @abstractmethod
    def extract_position(
        self, position_keys: Sequence[str], model_state: ModelState
    ) -> Position:
        """
        Extracts the position from the model state given a sequence of position keys.

        The returned position must have the same order as the supplied key sequence.
        """

        raise NotImplementedError

    @abstractmethod
    def update_state(self, position: Position, model_state: ModelState) -> ModelState:
        """Updates the model state with the values in the position."""

        raise NotImplementedError

    @abstractmethod
    def log_prob(self, model_state: ModelState) -> float:
        """Computes the unnormalized log-probability given the model state."""

        raise NotImplementedError


class Kernel(Protocol[TKernelState, TTransitionInfo, TTuningInfo]):
    """Protocol for a transition kernel."""

    error_book: ClassVar[dict[int, str]]
    """
    Maps error codes to error messages.

    We use -1 if an error code is required but the kernel is skipped.
    """

    needs_history: ClassVar[bool] = False
    """Is set to true if the kernel expects the history for tuning."""

    position_keys: tuple[str, ...]
    """Keys for which the kernel handles the transition."""

    identifier: str = ""
    """
    An identifier for the kernel object that is set by the `EngineBuilder`
    if it is an empty string.
    """

    @abstractmethod
    def set_model(self, model: ModelInterface):
        """Sets the model interface."""
        ...

    @abstractmethod
    def has_model(self) -> bool:
        """Whether the model interface is set."""
        ...

    @abstractmethod
    def init_state(self, prng_key: KeyArray, model_state: ModelState) -> KernelState:
        """Creates the initial kernel state."""

        raise NotImplementedError

    @abstractmethod
    def transition(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[TKernelState, TTransitionInfo]:
        """Handles one transition. Must be jittable."""

        raise NotImplementedError

    @abstractmethod
    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[TKernelState, TTuningInfo]:
        """
        The method can perform automatic tuning of the kernel and is called
        after each adaptation epoch.

        To tune the kernel, the method can return an altered kernel state.

        Must be jittable.

        Parameters
        ----------
        prng_key
            The key for JAX' pseudo-random number generator.
        model_state
            Current model state.
        kernel_state
            Current kernel state.
        epoch
            Current epoch state.
        history
            Holds the history of the position of the current epoch, i.e., that
            is the position but each leave in the pytree is enhanced by one
            dimension (axis = 0) which represents the time or MCMC iteration.
            The value may be ``None`` if the class variable :py:attr:`~needs_histroy` is
            ``False``.
        """

        raise NotImplementedError

    @abstractmethod
    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TKernelState:
        """
        Called at the beginning of an epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TKernelState:
        """
        Called at the end of an epoch. Must be jittable.
        """

        raise NotImplementedError

    @abstractmethod
    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: TKernelState,
        model_state: ModelState,
        tuning_history: TTuningInfo | None,
    ) -> WarmupOutcome[TKernelState]:
        """
        Asks the kernel to inspect the warmup history and react to it.

        This method is executed once the first non-warmup epoch is encountered and
        before :meth:`.start_epoch` is called.

        ``tuning_history`` is ``None`` if no tuning has happened prior to the first
        non-warmup epoch. Otherwise ``tuning_history`` has the same structure as
        returned from :meth:`.tune` but each leaf has an additional dimension. The first
        dimension refers to the n-the tuning.

        Must be jittable.

        The signature is likely to change.
        """

        raise NotImplementedError


class GeneratedQuantity(Protocol):
    """
    Protocol representing the data structure for quantities generated via
    implmentations of :class:`.QuantityGenerator`.

    Concrete implementations should add additional attributes.

    The attribute ``error_code`` is reserved to store integers that map to error
    messages via the error book provided in the implmentation of
    :class:`.QuantityGenerator`.
    """

    error_code: int


TGeneratedQuantity = TypeVar(
    "TGeneratedQuantity", bound=GeneratedQuantity, covariant=True
)


class QuantityGenerator(Protocol[TGeneratedQuantity]):
    """
    Protocol for a class that calculates a quantity based on the model
    state and a random seed.
    """

    error_book: ClassVar[dict[int, str]]
    identifier: str

    @abstractmethod
    def set_model(self, model: ModelInterface):
        """
        Sets a model.

        Parameters
        ----------
        model
            The model to be set.
        """
        raise NotImplementedError

    @abstractmethod
    def has_model(self) -> bool:
        """*True*, if a model has been set with :meth:`.set_model`."""
        raise NotImplementedError

    def generate(
        self,
        prng_key: KeyArray,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TGeneratedQuantity:
        """
        Generates a new quantity based on the model and PRNG state.

        Parameters
        ----------
        prng_key
            The key for JAX' pseudo-random number generator.
        model_state
            Current model state.
        epoch
            Current epoch state.
        """



import jax
import pytest

from liesel.goose.epoch import EpochConfig, EpochManager, EpochType


def test_epoch_state_time() -> None:
    state = EpochConfig(EpochType.POSTERIOR, 10, 1, None).to_state(0, 0)

    assert state.time_left() == 10

    state.advance_time(5)
    assert state.time_left() == 5
    assert state.time == 5
    assert state.time_in_epoch == 5

    state.advance_time(5)
    assert state.time_left() == 0
    assert state.time == 10
    assert state.time_in_epoch == 10

    state = EpochConfig(EpochType.POSTERIOR, 10, 1, None).to_state(1, 10)
    assert state.time_left() == 10
    assert state.time == 10
    assert state.time_in_epoch == 0
    state.advance_time(10)
    assert state.time == 20
    assert state.time_in_epoch == 10
    assert state.time_left() == 0


def test_is_epoch_type() -> None:
    type = EpochType.INITIAL_VALUES
    assert not EpochType.is_adaptation(type)
    assert not EpochType.is_warmup(type)

    type = EpochType.SLOW_ADAPTATION
    assert EpochType.is_adaptation(type)
    assert EpochType.is_warmup(type)

    type = EpochType.FAST_ADAPTATION
    assert EpochType.is_adaptation(type)
    assert EpochType.is_warmup(type)

    type = EpochType.BURNIN
    assert not EpochType.is_adaptation(type)
    assert EpochType.is_warmup(type)

    type = EpochType.POSTERIOR
    assert not EpochType.is_adaptation(type)
    assert not EpochType.is_warmup(type)


def test_is_epoch_type_jitted() -> None:
    jis_adapt = jax.jit(EpochType.is_adaptation)
    jis_warmup = jax.jit(EpochType.is_warmup)
    type = EpochType.INITIAL_VALUES
    assert not jis_adapt(type)
    assert not jis_warmup(type)

    type = EpochType.SLOW_ADAPTATION
    assert jis_adapt(type)
    assert jis_warmup(type)

    type = EpochType.FAST_ADAPTATION
    assert jis_adapt(type)
    assert jis_warmup(type)

    type = EpochType.BURNIN
    assert not jis_adapt(type)
    assert jis_warmup(type)

    type = EpochType.POSTERIOR
    assert not jis_adapt(type)
    assert not jis_warmup(type)


def test_epoch_manager() -> None:
    configs = (
        EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None),
        EpochConfig(EpochType.FAST_ADAPTATION, 3, 1, None),
        EpochConfig(EpochType.POSTERIOR, 5, 1, None),
    )

    manager = EpochManager(configs)

    # in the beginning we are in no epoch
    assert manager.has_more()

    # advance to first epoch
    state = manager.next()
    assert state.nth_epoch == 0
    assert state.time_before_epoch == 0
    assert manager.has_more()

    # advance to second epoch
    state = manager.next()
    assert state.nth_epoch == 1
    assert state.time_before_epoch == 1
    assert manager.has_more()

    # advance to third epoch
    state = manager.next()
    assert state.nth_epoch == 2
    assert state.time_before_epoch == 4

    # manager is now exhausted
    assert not manager.has_more()
    with pytest.raises(RuntimeError):
        manager.next()

    # add epoch configs
    manager.append(EpochConfig(EpochType.POSTERIOR, 7, 1, None))
    manager.append(EpochConfig(EpochType.POSTERIOR, 11, 1, None))

    # more epoch are now available
    assert manager.has_more()

    # advance to fourth epoch
    state = manager.next()
    assert state.nth_epoch == 3
    assert state.time_before_epoch == 9

    # advance to fith epoch
    state = manager.next()
    assert state.nth_epoch == 4
    assert state.time_before_epoch == 16

    # assert that the config matches
    assert state.config.type == EpochType.POSTERIOR
    assert state.config.duration == 11

    # manager is now exhausted
    assert not manager.has_more()
    with pytest.raises(RuntimeError):
        manager.next()


def test_epoch_manager_with_faulty_configs() -> None:
    # first epoch must be initial values
    with pytest.raises(RuntimeError):
        _ = EpochManager([EpochConfig(EpochType.POSTERIOR, 1, 1, None)])

    # first epoch must have length 1
    with pytest.raises(RuntimeError):
        _ = EpochManager([EpochConfig(EpochType.INITIAL_VALUES, 2, 1, None)])

    manager = EpochManager([EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None)])

    # initial values may not follow at a later position
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None))

    # thinning must be smaller or equal to duration
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.FAST_ADAPTATION, 5, 10, None))

    # warm up cannot follow posterior
    manager.append(EpochConfig(EpochType.FAST_ADAPTATION, 3, 1, None))
    manager.append(EpochConfig(EpochType.POSTERIOR, 4, 1, None))
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.FAST_ADAPTATION, 1, 1, None))

    # duration must be positive
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.POSTERIOR, -1, 1, None))

    # duration must be multiple of thinning
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.POSTERIOR, 13, 7, None))


def test_epoch_manager_with_thinning() -> None:
    manager = EpochManager([EpochConfig(EpochType.INITIAL_VALUES, 1, 1, None)])

    # thinning must be positive
    with pytest.raises(RuntimeError):
        manager.append(EpochConfig(EpochType.FAST_ADAPTATION, 20, -1, None))

    # thinning can be larger than one in WARMUP
    manager.append(EpochConfig(EpochType.FAST_ADAPTATION, 100, 10, None))

    # thinning can be larger than one in POSTERIOR epoch
    manager.append(EpochConfig(EpochType.POSTERIOR, 100, 20, None))



"""contains a deterministic kernel used for test purposes"""

from collections.abc import Sequence
from dataclasses import dataclass
from typing import ClassVar

import jax.numpy as jnp

from liesel.goose.epoch import EpochState, EpochType
from liesel.goose.kernel import (
    DefaultTransitionInfo,
    DefaultTuningInfo,
    TransitionMixin,
    TransitionOutcome,
    TuningOutcome,
    WarmupOutcome,
)
from liesel.goose.pytree import register_dataclass_as_pytree
from liesel.goose.types import KeyArray, ModelInterface, ModelState, Position


@register_dataclass_as_pytree
@dataclass
class DetCountingKernelState:
    increment_per_transition: int
    total_transitions: int
    total_adaptive_transitions: int
    post_warmup_transitions: int
    epoch_counter: int
    in_epoch: bool
    tune_counter: int
    warmup_finialized: bool

    @staticmethod
    def default() -> "DetCountingKernelState":
        return DetCountingKernelState(
            increment_per_transition=1,
            total_transitions=0,
            total_adaptive_transitions=0,
            post_warmup_transitions=0,
            epoch_counter=0,
            in_epoch=False,
            tune_counter=0,
            warmup_finialized=False,
        )


@register_dataclass_as_pytree
@dataclass
class DetCountingKernelTuningInfo(DefaultTuningInfo):
    pass


@register_dataclass_as_pytree
@dataclass
class DetCountingTransInfo(DefaultTransitionInfo):
    useless_field: float = 1.0


class DetCountingKernel(TransitionMixin[DetCountingKernelState, DetCountingTransInfo]):
    error_book: ClassVar[dict[int, str]] = {0: "no errors"}
    needs_history: ClassVar[bool] = False
    identifier: str = ""

    def __init__(
        self,
        position_keys: Sequence[str],
        initial_state: DetCountingKernelState,
    ):
        self._model: None | ModelInterface = None
        self.position_keys = tuple(position_keys)
        self._initial_state = initial_state

    @property
    def model(self) -> ModelInterface:
        """Returns the model interface if it is set. Raises error otherwise."""

        if self._model is None:
            raise RuntimeError("Model interface not set")

        return self._model

    def set_model(self, model: ModelInterface):
        self._model = model

    def has_model(self) -> bool:
        return self._model is not None

    def init_state(
        self, prng_key: KeyArray, model_state: ModelState
    ) -> DetCountingKernelState:
        return self._initial_state

    def start_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> DetCountingKernelState:
        kernel_state.epoch_counter += 1
        kernel_state.in_epoch = True

        return kernel_state

    def end_epoch(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> DetCountingKernelState:

        kernel_state.in_epoch = False
        return kernel_state

    def _standard_transition(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[DetCountingKernelState, DetCountingTransInfo]:
        kernel_state.total_transitions += 1
        kernel_state.post_warmup_transitions += 1 * (
            1 - EpochType.is_warmup(epoch.config.type)
        )
        position = self.model.extract_position(self.position_keys, model_state)

        epoch_value = kernel_state.epoch_counter * 10000
        transition_value = kernel_state.increment_per_transition * epoch.time_in_epoch
        for pkey in position.keys():
            position[pkey] = jnp.asarray(epoch_value + transition_value)

        new_model_state = self.model.update_state(position, model_state)

        info = DetCountingTransInfo(0, 1.0, 1)
        return TransitionOutcome(info, kernel_state, new_model_state)

    def _adaptive_transition(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        epoch: EpochState,
    ) -> TransitionOutcome[DetCountingKernelState, DetCountingTransInfo]:
        outcome = self._standard_transition(prng_key, kernel_state, model_state, epoch)
        outcome.kernel_state.total_adaptive_transitions += 1
        return outcome

    def tune(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        epoch: EpochState,
        history: Position | None,
    ) -> TuningOutcome[DetCountingKernelState, DetCountingKernelTuningInfo]:
        kernel_state.tune_counter += 1
        info = DetCountingKernelTuningInfo(error_code=0, time=epoch.time)
        return TuningOutcome(info, kernel_state)

    def end_warmup(
        self,
        prng_key: KeyArray,
        kernel_state: DetCountingKernelState,
        model_state: ModelState,
        tuning_history: DetCountingKernelTuningInfo | None,
    ) -> WarmupOutcome[DetCountingKernelState]:
        kernel_state.warmup_finialized = True
        return WarmupOutcome(0, kernel_state)



"""
Liesel modeling framework.
"""

from .distreg import DistRegBuilder, dist_reg_mcmc, tau2_gibbs_kernel
from .goose import GooseModel
from .legacy import (
    PIT,
    Addition,
    Bijector,
    ColumnStack,
    DesignMatrix,
    Hyperparameter,
    InverseLink,
    Parameter,
    Predictor,
    RegressionCoef,
    Response,
    Smooth,
    SmoothingParam,
)
from .model import GraphBuilder, Model, load_model, save_model
from .nodes import (  # TODO: Bijector?
    Array,
    Calc,
    Data,
    Dist,
    Distribution,
    Group,
    InputGroup,
    Node,
    NodeState,
    Obs,
    Param,
    TransientCalc,
    TransientDist,
    TransientIdentity,
    TransientNode,
    Var,
    add_group,
)
from .viz import plot_nodes, plot_vars



"""
Liesel probabilistic programming framework.
"""

from .__version__ import __version__, __version_info__  # isort: skip

from . import bijectors, distributions, goose, model
from .logging import reset_logger, setup_logger

# because logger setup takes place after importing the submodules, it only affects
# log messages emitted at runtime
setup_logger()



import jax.numpy as jnp

from liesel.goose.pytree import concatenate_leaves, slice_leaves, stack_leaves


def test_slice():
    tree = {
        "foo": jnp.zeros((3, 3, 1)),
        "bar": (jnp.ones((3, 1)), jnp.arange(18).reshape((3, 3, 2))),
    }
    slobj = jnp.s_[:, ..., 0]
    nt = slice_leaves(tree, slobj)
    assert nt["foo"].shape == (3, 3)
    assert nt["bar"][0].shape == (3,)
    assert nt["bar"][1].shape == (3, 3)

    assert jnp.all(tree["foo"][slobj] == nt["foo"])


def test_stack():
    tree = {
        "foo": jnp.zeros((3, 3, 1)),
        "bar": (jnp.ones((3, 1)), jnp.arange(18).reshape((3, 3, 2))),
    }

    nt = stack_leaves([tree, tree])

    assert nt["foo"].shape == (2, 3, 3, 1)
    assert nt["bar"][0].shape == (2, 3, 1)
    assert nt["bar"][1].shape == (2, 3, 3, 2)
    assert jnp.all(jnp.stack([tree["foo"], tree["foo"]]) == nt["foo"])

    nt = stack_leaves([tree, tree], axis=1)

    assert nt["foo"].shape == (3, 2, 3, 1)
    assert nt["bar"][0].shape == (3, 2, 1)
    assert nt["bar"][1].shape == (3, 2, 3, 2)
    assert jnp.all(jnp.stack([tree["foo"], tree["foo"]], axis=1) == nt["foo"])


def test_concatenate():
    tree = {
        "foo": jnp.zeros((3, 3, 1)),
        "bar": (jnp.ones((3, 1)), jnp.arange(18).reshape((3, 3, 2))),
    }

    nt = concatenate_leaves([tree, tree])

    assert nt["foo"].shape == (6, 3, 1)
    assert nt["bar"][0].shape == (6, 1)
    assert nt["bar"][1].shape == (6, 3, 2)
    assert jnp.all(jnp.concatenate([tree["foo"], tree["foo"]]) == nt["foo"])

    nt = concatenate_leaves([tree, tree], axis=1)

    assert nt["foo"].shape == (3, 6, 1)
    assert nt["bar"][0].shape == (3, 2)
    assert nt["bar"][1].shape == (3, 6, 2)
    assert jnp.all(jnp.concatenate([tree["foo"], tree["foo"]], axis=1) == nt["foo"])



"""
Logging utilities.
"""

import logging
from pathlib import Path


def setup_logger() -> None:
    """
    Sets up a basic ``StreamHandler`` that prints log messages to the terminal.
    The default log level of the ``StreamHandler`` is set to "info".

    The global log level for Liesel can be adjusted like this::

        import logging
        logger = logging.getLogger("liesel")
        logger.level = logging.WARNING

    This will set the log level to "warning".
    """

    # We adjust only our library's logger
    logger = logging.getLogger("liesel")

    # This is the level that will in principle be handled by the logger.
    # If it is set, for example, to logging.WARNING, this logger will never
    # emit messages of a level below warning
    logger.setLevel(logging.DEBUG)

    # By setting this to False, we prevent the Liesel log messages from being passed on
    # to the root logger. This prevents duplication of the log messages
    logger.propagate = False

    # This is the default handler that we set for our log messages
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)

    # We define the format of log messages for this handler
    formatter = logging.Formatter("%(name)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)

    logger.addHandler(handler)


def reset_logger() -> None:
    """
    Resets the Liesel logger.

    Specifically, this function...

    - ... resets the level of the Liesel logger to ``logging.NOTSET``.
    - ... sets ``propagate=True`` for the Liesel logger.
    - ... removes *all* handlers from the Liesel logger.

    This function is useful if you want to set up a custom logging configuration.
    """

    # We adjust only our library's logger
    logger = logging.getLogger("liesel")

    # Removes the level of the logger. All log messages will be propagated
    logger.setLevel(logging.NOTSET)

    # By setting this to True, we allow the Liesel log messages to be passed on
    # to the root logger
    logger.propagate = True

    # Removes all handlers from the Liesel logger
    for handler in logger.handlers:
        logger.removeHandler(handler)


def add_file_handler(
    path: str | Path,
    level: str,
    logger: str = "liesel",
    fmt: str = "%(asctime)s - %(levelname)s - %(name)s - %(message)s",
) -> None:
    """
    Adds a file handler to a logger.

    Parameters
    ----------
    path
        Absolute path to the log file. If it does not exist, it will be created.
        If any parent directory does not exist, it will be created as well.
    level
        The log level of the messages to write to the file. Can be ``"debug"``,
        ``"info"``, ``"warning"``, ``"error"`` or ``"critical"``. The file will
        contain all messages from the specified level upwards.
    logger
        The name of the logger to configure the file handler for. Can be, for example,
        the full name of a Liesel module. For :mod:`liesel.goose`, the argument should
        be specified as ``"liesel.goose"``, etc.
    fmt
        Formatting string. See the documentation of the :class:`logging.Formatter`.

    Examples
    --------
    A basic file handler catching all log messages from Liesel::

        import logging
        import liesel as lsl

        lsl.logging.add_file_handler(path="path/to/logfile.log", level="debug")

        logger = logging.getLogger("liesel")
        logger.warning("My warning message")

    A file handler that catches only log messages from the :mod:`liesel.goose` module
    of level "warning" or higher::

        import logging
        import liesel as lsl

        lsl.logging.add_file_handler(
            path="path/to/goose_warnings.log",
            level="warning",
            logger="liesel.goose"
        )

        logger = logging.getLogger("liesel")
        logger.warning("My warning message")
    """

    path = Path(path)

    if not path.is_absolute():
        raise ValueError("Provided path for logging file handler must be absolute")

    path.parent.mkdir(parents=True, exist_ok=True)

    _logger = logging.getLogger(logger)
    handler = logging.FileHandler(path)

    handler.setLevel(getattr(logging, level.upper()))

    formatter = logging.Formatter(fmt, "%Y-%m-%d %H:%M:%S")
    handler.setFormatter(formatter)

    _logger.addHandler(handler)



"""
Nodes and variables.
"""

from __future__ import annotations

import weakref
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable
from functools import wraps
from itertools import chain
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, Union

import tensorflow_probability.substrates.jax.bijectors as jb
import tensorflow_probability.substrates.jax.distributions as jd
import tensorflow_probability.substrates.numpy.bijectors as nb
import tensorflow_probability.substrates.numpy.distributions as nd
from deprecated.sphinx import deprecated

from ..distributions.nodist import NoDistribution

if TYPE_CHECKING:
    from .model import Model

__all__ = [
    "Array",
    "Bijector",
    "Calc",
    "Data",
    "Dist",
    "Distribution",
    "Group",
    "InputGroup",
    "Node",
    "NodeState",
    "Obs",
    "Param",
    "TransientCalc",
    "TransientDist",
    "TransientIdentity",
    "TransientNode",
    "Var",
    "add_group",
]

Array = Any
Distribution = Union[jd.Distribution, nd.Distribution]
Bijector = Union[jb.Bijector, nb.Bijector]

T = TypeVar("T", bound=Hashable)


def _unique_tuple(*args: Iterable[T]) -> tuple[T, ...]:
    return tuple(dict.fromkeys(chain(*args)))


def in_model_method(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        if not self.model:
            raise RuntimeError(
                f"{repr(self)} is not part of a model, cannot call {fn.__name__}()"
            )
        return fn(self, *args, **kwargs)

    return wrapped


def in_model_getter(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        if not self.model:
            raise RuntimeError(
                f"{repr(self)} is not part of a model, cannot call '{fn.__name__}'"
            )
        return fn(self, *args, **kwargs)

    return wrapped


def no_model_method(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        if self.model:
            raise RuntimeError(
                f"{repr(self)} is part of a model, cannot call {fn.__name__}()"
            )
        return fn(self, *args, **kwargs)

    return wrapped


def no_model_setter(fn):
    @wraps(fn)
    def wrapped(self, *args, **kwargs):
        if self.model:
            raise RuntimeError(
                f"{repr(self)} is part of a model, cannot set '{fn.__name__}'"
            )
        return fn(self, *args, **kwargs)

    return wrapped


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Nodes ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class NodeState(NamedTuple):
    """The state of a node."""

    value: Any
    """The value of the node."""

    outdated: bool
    """Whether the node is outdated."""

    extra: Any = None
    """Optional extra information."""


class Node(ABC):
    """
    A node of a computational graph that can cache its value.

    Liesel represents statistical models as directed acyclic graphs (DAGs) of
    random variables (see :class:`.Var`) and computational nodes. The graph of
    random variables is built on top of the computational graph. The nodes of the
    computational graph will typically express computations in JAX returning arrays
    or pytrees_, but in general, they can represent arbitrary operations in Python.

    Nodes can cache the result of the operations they represent, improving
    the efficiency of the graph. The cached values are part of the model state
    (see :attr:`.Model.state`), and can be stored in a chain by Liesel's MCMC engine
    Goose.

    This class is an abstract class that cannot be initialized without defining the
    :meth:`.update` method. See below for the most important concrete node classes.

    Additional meta-information can be added by wrapping a :class:`.Data` or
    :class:`.Calc` node in a :class:`.Var`, which is particularly useful if the node
    represents a random variable, but can also be a good idea in a few other
    situations.

    See Also
    --------
    .Calc :
        A node representing a general calculation/operation
        in JAX or Python.
    .Data :
        A node representing some static data.
    .Dist :
        A node representing a `TensorFlow Probability`_ distribution,
        typically for the evaluation of the log-probability.


    .. _pytrees: https://jax.readthedocs.io/en/latest/pytrees.html
    .. _TensorFlow Probability: https://www.tensorflow.org/probability
    """

    def __init__(
        self,
        *inputs: Any,
        _name: str = "",
        _needs_seed: bool = False,
        **kwinputs: Any,
    ):
        self._groups: dict[str, Group] = {}
        self._inputs = tuple(self._to_node(_input) for _input in inputs)
        self._kwinputs = {kw: self._to_node(_input) for kw, _input in kwinputs.items()}
        self._model: weakref.ref[Model] | Callable[[], None] = lambda: None
        self._name = _name
        self._needs_seed = _needs_seed
        self._outdated = True
        self._outputs: tuple[Node, ...] = ()
        self._value: Any = None
        self._var: Var | None = None

        self.monitor = False
        """Whether the node should be monitored by an inference algorithm."""

    def _add_output(self, output: Node) -> Node:
        self._outputs = _unique_tuple(self._outputs, [output])
        return self

    def _clear_outputs(self) -> Node:
        self._outputs = ()
        return self

    def _set_model(self, model: Model) -> Node:
        if self.model:
            raise RuntimeError(f"{repr(self)} can only be part of one model")

        self._model = weakref.ref(model)
        return self

    def _set_var(self, var: Var) -> Node:
        if self.var:
            raise RuntimeError(f"{repr(self)} can only be part of one var")

        self._var = var
        return self

    @staticmethod
    def _to_node(x: Any) -> Node:
        if isinstance(x, Var):
            return x.var_value_node

        if not isinstance(x, Node):
            return Data(x)

        return x

    def _unset_model(self) -> Node:
        self._model = lambda: None
        return self

    def _unset_var(self) -> Node:
        self._var = None
        return self

    def all_input_nodes(self) -> tuple[Node, ...]:
        """Returns all non-keyword and keyword input nodes as a unique tuple."""
        return _unique_tuple(self.inputs, self.kwinputs.values())

    @in_model_method
    def all_output_nodes(self) -> tuple[Node, ...]:
        """Returns all output nodes as a unique tuple."""
        return self.outputs

    def clear_state(self) -> Node:
        """Clears the state of the node."""
        self.state = NodeState(None, True)
        return self

    @in_model_method
    def flag_outdated(self) -> Node:
        """Flags the node and its recursive outputs as outdated."""
        self._outdated = True

        for node in self._outputs:
            node.flag_outdated()

        return self

    @property
    def groups(self) -> MappingProxyType[str, Group]:
        """The groups that this node is a part of."""
        return MappingProxyType(self._groups)

    @property
    def inputs(self) -> tuple[Node, ...]:
        """The non-keyword input nodes."""
        return self._inputs

    @property
    def kwinputs(self) -> MappingProxyType[str, Node]:
        """The keyword input nodes."""
        return MappingProxyType(self._kwinputs)

    @property
    def model(self) -> Model | None:
        """The model the node is part of."""
        return self._model()

    @property
    def name(self) -> str:
        """The name of the node."""
        return self._name

    @name.setter
    @no_model_setter
    def name(self, name: str):
        self._name = name

    @property
    def needs_seed(self) -> bool:
        """Whether the node needs a seed / PRNG key."""
        return self._needs_seed

    @needs_seed.setter
    @no_model_setter
    def needs_seed(self, needs_seed: bool):
        self._needs_seed = needs_seed

    @property
    def outdated(self) -> bool:
        """Whether the node is outdated."""

        if not self.model:
            return True

        return self._outdated

    @property
    @in_model_getter
    def outputs(self) -> tuple[Node, ...]:
        """The output nodes."""
        return self._outputs

    @no_model_method
    def set_inputs(self, *inputs: Any, **kwinputs: Any) -> Node:
        """Sets the non-keyword and keyword input nodes."""
        self._inputs = tuple(self._to_node(_input) for _input in inputs)

        self._kwinputs.clear()
        kwinputs = {kw: self._to_node(_input) for kw, _input in kwinputs.items()}
        self._kwinputs.update(kwinputs)

        return self

    @property
    def state(self) -> NodeState:
        """
        The state of the node.

        For the default node, a :class:`.NodeState` with the value and the
        outdated flag, but subclasses can add extra information to the state.
        """
        return NodeState(self.value, self.outdated)

    @state.setter
    def state(self, state: NodeState):
        self._value = state.value
        self._outdated = state.outdated

    @abstractmethod
    def update(self) -> Node:
        """Updates the value of the node."""

    @property
    def value(self) -> Any:
        """
        The value of the node.

        Can only be set for a :class:`.Data` node, but not a :class:`.Calc` or
        :class:`.Dist` node. If the node is part of a :class:`.Model` ``m`` with
        ``m.auto_update == True``, setting the value of the node triggers an update
        of the model. The auto-update can be disabled to improve the performance if
        multiple model parameters are updated at once.
        """
        return self._value

    @property
    def var(self) -> Var | None:
        """The variable the node is part of."""
        return self._var

    def __getstate__(self):
        state = self.__dict__.copy()
        state["_model"] = self._model()
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

        if self._model is not None:
            self._model = weakref.ref(self._model)
        else:
            self._model = lambda: None

    def __repr__(self) -> str:
        return f'{type(self).__name__}(name="{self.name}")'


class TransientNode(Node):
    """
    A node that does not cache its value.

    A transient node is outdated if and only if at least one of its input nodes
    is outdated. The :attr:`.outdated` property checks this condition on-the-fly.
    """

    @property
    def outdated(self) -> bool:
        """
        Whether the node is outdated.

        A transient node is outdated if and only if at least one of its input nodes
        is outdated. This condition is checked on-the-fly.
        """

        if not self.model:
            return True

        return any(_input.outdated for _input in self.all_input_nodes())

    @property
    def state(self) -> NodeState:
        """The state of the node with the value ``None``."""
        return NodeState(None, self.outdated)

    @state.setter
    def state(self, state: NodeState):
        self._value = state.value
        self._outdated = state.outdated

    def update(self):
        """Does nothing."""
        return self

    @property
    @abstractmethod
    def value(self) -> Any:
        """
        The value of the node.

        Computed on-the-fly.
        """


class ArgGroup(NamedTuple):
    """A group of arguments as a named tuple of ``args`` and ``kwargs``."""

    args: list[Any]
    """The non-keyword arguments."""

    kwargs: dict[str, Any]
    """The keyword arguments."""


class InputGroup(TransientNode):
    """
    A node that groups its inputs for another node.

    Essentially, this node "forwards" the values of its inputs to its outputs
    as an :class:`.ArgGroup`.
    """

    @property
    def value(self) -> ArgGroup:
        args = [_input.value for _input in self.inputs]
        kwargs = {kw: _input.value for kw, _input in self.kwinputs.items()}
        return ArgGroup(args, kwargs)


class Data(Node):
    """A data node. Always up-to-date."""

    def __init__(self, value: Any, _name: str = ""):
        super().__init__(_name=_name)
        self._value = value

    def flag_outdated(self) -> Data:
        """Stops the recursion setting outdated flags."""
        return self

    @property
    def outdated(self) -> bool:
        return False

    def update(self) -> Data:
        """Does nothing."""
        return self

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any):
        self._value = value

        if self.model:
            for node in self.outputs:
                node.flag_outdated()

            if self.model.auto_update:
                self.model.update()


class Calc(Node):
    """A calculator node."""

    def __init__(
        self,
        function: Callable[..., Any],
        *inputs: Any,
        _name: str = "",
        _needs_seed: bool = False,
        **kwinputs: Any,
    ):
        super().__init__(*inputs, **kwinputs, _name=_name, _needs_seed=_needs_seed)
        self._function = function

    @property
    def function(self) -> Callable[..., Any]:
        """The wrapped function."""
        return self._function

    @function.setter
    @no_model_setter
    def function(self, function: Callable[..., Any]):
        self._function = function

    def update(self) -> Calc:
        args = [_input.value for _input in self.inputs]
        kwargs = {kw: _input.value for kw, _input in self.kwinputs.items()}
        self._value = self.function(*args, **kwargs)
        self._outdated = False
        return self


class TransientCalc(TransientNode, Calc):
    """A transient calculator node that does not cache its value."""

    @property
    def value(self) -> Any:
        args = [_input.value for _input in self.inputs]
        kwargs = {kw: _input.value for kw, _input in self.kwinputs.items()}
        return self.function(*args, **kwargs)


class TransientIdentity(TransientCalc):
    """
    A transient identity node that does not cache its value.

    Essentially, this node "forwards" the value of its input to its outputs.
    """

    def __init__(self, _input: Any, _name: str = ""):
        super().__init__(lambda x: x, _input, _name=_name)


class VarValue(TransientIdentity):
    """
    A proxy node for the value of a :class:`.Var`.

    This node type is used to keep the references to a variable intact,
    even if the underlying value node is replaced.
    """


class Dist(Node):
    """A distribution node."""

    def __init__(
        self,
        distribution: Callable[..., Distribution],
        *inputs: Any,
        _name: str = "",
        _needs_seed: bool = False,
        **kwinputs: Any,
    ):
        super().__init__(*inputs, **kwinputs, _name=_name, _needs_seed=_needs_seed)

        self._at: Node | None = None
        self._distribution = distribution
        self._per_obs = True

    def all_input_nodes(self) -> tuple[Node, ...]:
        inputs = super().all_input_nodes()

        if self.at:
            inputs = _unique_tuple(inputs, [self.at])

        return inputs

    @property
    def at(self) -> Node | None:
        """Where to evaluate the distribution."""
        return self._at

    @at.setter
    @no_model_setter
    def at(self, at: Node | None):
        if self.var and at is not self.var.var_value_node:
            raise RuntimeError(
                f"{repr(self)} is part of a var, cannot set property `at`"
            )

        self._at = at

    @property
    def distribution(self) -> Callable[..., Distribution]:
        """The wrapped distribution."""
        return self._distribution

    @distribution.setter
    @no_model_setter
    def distribution(self, distribution: Callable[..., Distribution]):
        self._distribution = distribution

    def init_dist(self) -> Distribution:
        """Initializes the distribution."""
        args = [_input.value for _input in self.inputs]
        kwargs = {kw: _input.value for kw, _input in self.kwinputs.items()}
        dist = self.distribution(*args, **kwargs)
        return dist

    @property
    def log_prob(self) -> Array:
        """The log-probability of the distribution."""
        return self.value

    @property
    def per_obs(self) -> bool:
        """Whether the log-probability is stored per observation or summed up."""
        return self._per_obs

    @per_obs.setter
    @no_model_setter
    def per_obs(self, per_obs: bool):
        self._per_obs = per_obs

    def update(self) -> Dist:
        if not self.at:
            raise RuntimeError(
                f"{repr(self)} cannot evaluate log-prob, property `at` not set"
            )

        log_prob = self.init_dist().log_prob(self.at.value)

        if not self.per_obs and hasattr(log_prob, "sum"):
            log_prob = log_prob.sum()

        self._value = log_prob
        self._outdated = False
        return self


class TransientDist(TransientNode, Dist):
    """A transient distribution node that does not cache its value."""

    @property
    def value(self) -> Any:
        if not self.at:
            raise RuntimeError(
                f"{repr(self)} cannot evaluate log-prob, property `at` not set"
            )

        log_prob = self.init_dist().log_prob(self.at.value)

        if not self.per_obs and hasattr(log_prob, "sum"):
            log_prob = log_prob.sum()

        return log_prob


class NoDist(Dist):
    def __init__(self):
        super().__init__(NoDistribution)

    def all_input_nodes(self) -> tuple[Node, ...]:
        return ()

    def all_output_nodes(self) -> tuple[Node, ...]:
        return ()

    @property
    def outputs(self) -> tuple[Node, ...]:
        return ()

    def update(self) -> NoDist:
        return self

    @property
    def value(self) -> float:
        return 0.0


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Variable ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Var:
    """
    A variable in a statistical model, typically with a probability distribution.

    A variable in Liesel is typically a random variable, e.g. an observed or
    latent variable with a probability distribution, or a model parameter with
    a prior distribution. Note that observed variables and model parameters should
    typically be declared with the :func:`.Obs` and :func:`.Param` helper functions.
    Other quantities can also be declared as variables, e.g. fixed data like
    hyperparameters or design matrices, or quantities that are computed from
    other nodes, e.g. structured additive predictors in semi-parametric
    regression models.

    If a :class:`.Data` or :class:`.Calc` node does not have an associated probability
    distribution, it is possible but not necessary to declare it as a variable. There
    is no hard and fast rule when a node without a probability distribution should be
    declared as a variable and when not. The advantages of a variable in this case are:
    (1) easier access via the :attr:`.Model.vars` attribute, and (2) more explicit
    visualization with the :func:`.plot_vars` function. This might be particularly
    desirable for the hyperparameters of a prior distribution.

    See Also
    --------
    .Obs : Helper function to declare a variable as an observed quantity.
    .Param : Helper function to declare a variable as a model parameter.
    """

    __slots__ = (
        "_dist_node",
        "_groups",
        "_name",
        "_observed",
        "_parameter",
        "_role",
        "_value_node",
        "_var_value_node",
        "info",
    )

    def __init__(
        self,
        value: Any,
        distribution: Dist | None = None,
        name: str = "",
    ):
        self._name = name
        self._value_node: Node = Data(None)
        self._dist_node: Dist = NoDist()

        self._var_value_node: VarValue = VarValue(
            self._value_node, _name=f"{self._name}_var_value"
        )

        self._var_value_node._set_var(self)

        # use setters
        self.value_node = value  # type: ignore  # unfrozen
        self.dist_node = distribution  # type: ignore  # unfrozen

        self._observed = False
        self._parameter = False
        self._role = ""

        self._groups: dict[str, Group] = {}

        self.info: dict[str, Any] = {}
        """Additional meta-information about the variable as a dict."""

    def all_input_nodes(self) -> tuple[Node, ...]:
        """Returns all input *nodes* as a unique tuple."""
        inputs1 = self.value_node.all_input_nodes()
        inputs2 = self._dist_node.all_input_nodes()
        return _unique_tuple(inputs1, inputs2)

    def all_input_vars(self) -> tuple[Var, ...]:
        """
        Returns all input *variables* as a unique tuple.

        The returned tuple also contains input variables that are indirect inputs
        of this variable through nodes without variables.
        """
        nodes = list(self.all_input_nodes())
        visited = []
        _vars = []

        while nodes:
            node = nodes.pop()

            if node not in visited:
                if node.var and node.var is not self:
                    _vars.append(node.var)
                else:
                    nodes.extend(node.all_input_nodes())

                visited.append(node)

        return _unique_tuple(_vars)

    @in_model_method
    def all_output_nodes(self) -> tuple[Node, ...]:
        """Returns all output *nodes* as a unique tuple."""
        nodes = list(self.value_node.all_output_nodes())
        nodes.extend(self.var_value_node.all_output_nodes())
        nodes.extend(self._dist_node.all_output_nodes())
        nodes = [node for node in nodes if node is not self.var_value_node]
        return _unique_tuple(nodes)

    @in_model_method
    def all_output_vars(self) -> tuple[Var, ...]:
        """
        Returns all output *variables* as a unique tuple.

        The returned tuple also contains output variables that are indirect outputs
        of this variable through nodes without variables.
        """
        nodes = list(self.all_output_nodes())
        visited = []
        _vars = []

        while nodes:
            node = nodes.pop()

            if node not in visited:
                if node.var and node.var is not self:
                    _vars.append(node.var)
                else:
                    nodes.extend(node.all_output_nodes())

                visited.append(node)

        return _unique_tuple(_vars)

    @property
    def dist_node(self) -> Dist | None:
        """The distribution node of the variable."""
        return self._dist_node if self.has_dist else None

    @dist_node.setter
    @no_model_setter
    def dist_node(self, dist_node: Dist | None):
        if not dist_node:
            dist_node = NoDist()

        if dist_node.model:
            raise RuntimeError(
                f"{repr(dist_node)} is part of a model, cannot be set as dist node"
            )

        if self.name and not dist_node.name:
            dist_node.name = f"{self.name}_log_prob"  # type: ignore  # unfrozen

        self._dist_node._unset_var()

        dist_node._set_var(self)
        dist_node.at = self.var_value_node  # type: ignore  # unfrozen
        self._dist_node = dist_node

    @property
    def groups(self) -> MappingProxyType[str, Group]:
        """The groups that this variable is a part of."""
        return MappingProxyType(self._groups)

    @property
    def has_dist(self) -> bool:
        """Whether the variable has a probability distribution."""
        return not isinstance(self._dist_node, NoDist)

    @property
    def log_prob(self) -> Array:
        """
        The log-probability of the variable.

        A variable without a probability distribution has a log-probability of 0.0.
        """
        return self._dist_node.value

    @property
    def model(self) -> Model | None:
        """The model the variable is part of."""
        return self.value_node.model

    @property
    def name(self) -> str:
        """The name of the variable."""
        return self._name

    @name.setter
    @no_model_setter
    def name(self, name: str):
        if name and self.value_node.name in ("", f"{self.name}_value"):
            self.value_node.name = f"{name}_value"  # type: ignore  # unfrozen
            self.var_value_node.name = f"{name}_var_value"  # type: ignore  # unfrozen

        if name and self._dist_node.name in ("", f"{self.name}_log_prob"):
            self._dist_node.name = f"{name}_log_prob"  # type: ignore  # unfrozen

        self._name = name

    @property
    def nodes(self) -> list[Node]:
        """The nodes of the variable as a list."""
        nodes = [self.value_node, self.var_value_node]

        if self.dist_node:
            nodes.append(self.dist_node)

        return nodes

    @property
    def observed(self) -> bool:
        """Whether the variable is observed."""
        return self._observed

    @observed.setter
    @no_model_setter
    def observed(self, observed: bool):
        self._observed = observed

    @property
    def parameter(self) -> bool:
        """Whether the variable is a parameter."""
        return self._parameter

    @parameter.setter
    @no_model_setter
    def parameter(self, parameter: bool):
        self._parameter = parameter

    @property
    def role(self) -> str:
        """The role of the variable."""
        return self._role

    @role.setter
    def role(self, role: str):
        self._role = role

    @property
    def strong(self) -> bool:
        """Whether the variable is strong."""
        return isinstance(self.value_node, Data)

    def update(self) -> Var:
        """Updates the variable."""
        self.value_node.update()
        self._dist_node.update()
        return self

    @property
    def value(self) -> Any:
        """
        The value of the variable.

        Can only be set if the variable is strong. If the variable is part of
        a :class:`.Model` ``m`` with ``m.auto_update == True``, setting the value of
        the variable triggers an update of the model. The auto-update can be disabled
        to improve the performance if multiple model parameters are updated at once.
        """
        return self.value_node.value

    @value.setter
    def value(self, value: Any):
        if self.weak:
            raise RuntimeError(f"{repr(self)} is weak, cannot set value")

        self.value_node.value = value  # type: ignore  # data node

    @property
    def value_node(self) -> Node:
        """The value node of the variable."""
        return self._value_node

    @value_node.setter
    @no_model_setter
    def value_node(self, value_node: Any):
        if isinstance(value_node, Var):
            value_node = Calc(lambda x: x, value_node)

        if not isinstance(value_node, Node):
            value_node = Data(value_node)

        if value_node.model:
            raise RuntimeError(
                f"{repr(value_node)} is part of a model, cannot be set as value node"
            )

        if self.name and not value_node.name:
            value_node.name = f"{self.name}_value"

        self.value_node._unset_var()

        value_node._set_var(self)
        self._value_node = value_node
        self._var_value_node.set_inputs(self._value_node)

    @property
    def var_value_node(self) -> VarValue:
        """The proxy node for the value of the variable."""
        return self._var_value_node

    @property
    def weak(self) -> bool:
        """Whether the variable is weak."""
        return not self.strong

    def __repr__(self) -> str:
        return f'{type(self).__name__}(name="{self.name}")'


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Variable helpers ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def Obs(value: Any | Calc, distribution: Dist | None = None, name: str = "") -> Var:
    """
    Declares an observed variable.

    Sets the :attr:`.Var.observed` flag. If the observed variable is a
    random variable, i.e. if it has an associated probability distribution,
    its log-probability is automatically added to the model log-likelihood
    (see :attr:`.Model.log_lik`).
    """
    var = Var(value, distribution, name)
    var.observed = True
    return var


def Param(value: Any | Calc, distribution: Dist | None = None, name: str = "") -> Var:
    """
    Declares a parameter variable.

    Sets the :attr:`.Var.parameter` flag. If the parameter variable is a
    random variable, i.e. if it has an associated probability distribution,
    its log-probability is automatically added to the model log-prior
    (see :attr:`.Model.log_prior`).
    """
    var = Var(value, distribution, name)
    var.value_node.monitor = True
    var.parameter = True
    return var


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Other helpers ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


@deprecated(reason="Use the new `Group` object directly", version="0.2.2")
def add_group(name: str, **kwargs: Node | Var) -> Group:
    """
    Assigns the nodes and variables to a group.

    See :attr:`liesel.model.nodes.Node.groups`.

    Parameters
    ----------
    name
        The name of the group.
    kwargs
        The nodes and variables in the group with their keys in the group
        as keywords.
    """
    return Group(name, **kwargs)


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Group ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Group:
    """
    A group holds a collection of related :class:`.Var` and/or :class:`.Node` objects.

    They allow you to do three basic things:

    1. Store related nodes together for easier access.
    2. Access their member nodes and variables via ``group["name"]``, where ``"name"``
       is the group-specific name, which can be different from the :attr:`.Var.name` /
       :attr:`.Node.name`.
    3. Easily retrieve a variable's or a node's value from a :attr:`.Model.state` based
       on their group-specific name via :meth:`.value_from`.

    Parameters
    ----------
    name
        The group's name. Must be unique among the groups of its members, and within
        a model.
    **nodes_and_vars
        An arbitrary number of nodes or variables. The keywords will be used as the
        group-specific names of the respective objects.

    See Also
    --------
    * :attr:`.Node.groups` and :attr:`.Var.groups` are :obj:`MappingProxyType` objects
      (basically read-only dictionaries) of the groups whose member the respective
      object is.
    * :meth:`.GraphBuilder.groups` and :meth:`.Model.groups` are methods that collect
      and return all groups within the graph/model.

    Notes
    -----
    Note the following:

    - Groups can only be filled upon initialization.
    - After initialization, variables and nodes cannot be removed from a group.

    Examples
    --------
    Add a variable to a group:

    >>> my_var = lsl.Var(0.0, name="long_unique_variable_name")
    >>> grp = lsl.Group(name="demo_group", short_name=my_var)
    >>> grp
    Group(name="demo_group")

    Access the variable by its group-specific name:

    >>> grp["short_name"]
    Var(name="long_unique_variable_name")

    Retrieve the value of a variable from a model state:

    >>> model_state = {my_var.value_node.name: lsl.NodeState(10., False)}
    >>> grp.value_from(model_state, "short_name")
    10.0
    """

    def __init__(self, name: str, **nodes_and_vars: Node | Var) -> None:
        self._name = name
        self._nodes_and_vars = nodes_and_vars

        for member in self._nodes_and_vars.values():
            if name in member.groups:
                raise RuntimeError(
                    f"{repr(member)} is already a member of a group "
                    f"with the name {repr(name)}"
                )
            member._groups[name] = self

        self._nodes = {
            name: obj
            for name, obj in self._nodes_and_vars.items()
            if isinstance(obj, Node)
        }

        self._vars = {
            name: obj
            for name, obj in self._nodes_and_vars.items()
            if isinstance(obj, Var)
        }

    @property
    def name(self) -> str:
        """The group's name."""
        return self._name

    def value_from(self, model_state: dict[str, NodeState], name: str) -> Array:
        """
        Retrieves the value of a node or variable that is a member of the group from
        a model state.

        Parameters
        ----------
        model_state
            The state of a Liesel model, i.e. a :class:`~.Model.state`.
        name
            The name of the node or variable within this group.

        Returns
        -------
        The value of the node or variable.
        """
        member = self[name]

        if isinstance(member, Var):
            value_name = member.value_node.name
        else:
            value_name = member.name

        return model_state[value_name].value

    @property
    def vars(self) -> MappingProxyType[str, Var]:
        """A mapping of the variables in the group with their names as keys."""
        return MappingProxyType(self._vars)

    @property
    def nodes(self) -> MappingProxyType[str, Node]:
        """A mapping of the nodes in the group with their names as keys."""
        return MappingProxyType(self._nodes)

    @property
    def nodes_and_vars(self) -> MappingProxyType[str, Node | Var]:
        """A mapping of all group members with their names as keys."""
        return MappingProxyType(self._nodes_and_vars)

    def __contains__(self, key) -> bool:
        return key in self._nodes_and_vars

    def __getitem__(self, key) -> Var | Node:
        return self._nodes_and_vars[key]

    def __repr__(self) -> str:
        return f'{type(self).__name__}(name="{self.name}")'



"""
Imitates the API from v0.1.
"""

from __future__ import annotations

from typing import Any, cast

import jax.numpy as jnp

from .nodes import Bijector as TFPBijector
from .nodes import Calc, Dist, Node, Obs, Param, Var

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Strong variables ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def DesignMatrix(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a design matrix."""
    var = Obs(value, distribution, name)
    var.role = "DesignMatrix"
    return var


def Hyperparameter(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a hyperparameter."""
    var = Var(value, distribution, name)
    var.role = "Hyperparameter"
    return var


def Parameter(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a model parameter."""
    var = Param(value, distribution, name)
    var.role = "Parameter"
    return var


def RegressionCoef(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a vector of regression coefficients."""
    var = Param(value, distribution, name)
    var.role = "RegressionCoef"
    return var


def Response(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a response vector."""
    var = Obs(value, distribution, name)
    var.role = "Response"
    return var


def SmoothingParam(
    value: Any | Calc, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A strong variable representing a smoothing parameter."""
    var = Param(value, distribution, name)
    var.role = "SmoothingParam"
    return var


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Weak variables ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def Addition(
    *inputs: Var | Node, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A weak variable calculating an element-wise sum."""
    calc = Calc(lambda *args, **kwargs: sum(args) + sum(kwargs.values()), *inputs)
    var = Var(calc, distribution, name)
    var.role = "Addition"
    return var


def Bijector(
    _input: Var | Node,
    bijector: type[TFPBijector],
    inverse: bool = False,
    distribution: Dist | None = None,
    name: str = "",
) -> Var:
    """
    A weak variable evaluating the ``forward()`` or ``inverse()`` method
    of a TFP bijector.
    """

    def fn(x):
        return bijector().forward(x) if not inverse else bijector().inverse(x)

    calc = Calc(fn, _input)
    var = Var(calc, distribution, name)
    var.role = "Bijector"
    return var


def ColumnStack(
    *inputs: Var | Node, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A weak variable stacking its inputs column-wise."""
    calc = Calc(lambda *args: jnp.column_stack(args), *inputs)
    var = Var(calc, distribution, name)
    var.role = "ColumnStack"
    return var


def InverseLink(
    _input: Var | Node,
    bijector: type[TFPBijector],
    inverse: bool = False,
    distribution: Dist | None = None,
    name: str = "",
) -> Var:
    """A weak variable representing an inverse link function."""
    var = Bijector(_input, bijector, inverse, distribution, name)
    var.role = "InverseLink"
    return var


class PITCalc(Node):
    """A probability integral transform (PIT) calculator node."""

    def __init__(
        self,
        _input: Dist,
        _name: str = "",
        _needs_seed: bool = False,
    ):
        super().__init__(_input, _name=_name, _needs_seed=_needs_seed)

    def update(self) -> PITCalc:
        dist = cast(Dist, self.inputs[0])

        if not dist.at:
            raise RuntimeError(
                f"Cannot evaluate PIT on {repr(dist)}, property `at` not set"
            )

        self._value = dist.init_dist().cdf(dist.at.value)
        self._outdated = False
        return self


def PIT(_input: Var | Dist, distribution: Dist | None = None, name: str = "") -> Var:
    """A weak variable evaluating a probability integral transform (PIT)."""
    dist = _input.dist_node if isinstance(_input, Var) else _input

    if not dist:
        raise RuntimeError(f"Cannot evaluate PIT on {repr(_input)}, has no dist node")

    if _input.name and not name:
        name = f"{_input.name}_pit"

    calc = PITCalc(dist)
    var = Var(calc, distribution, name)
    var.role = "PIT"
    return var


def Predictor(
    *inputs: Var | Node, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A weak variable representing a regression predictor."""
    var = Addition(*inputs, distribution=distribution, name=name)
    var.role = "Predictor"
    return var


def Smooth(
    x: Var | Node, beta: Var | Node, distribution: Dist | None = None, name: str = ""
) -> Var:
    """A weak variable calculating the matrix-vector product ``x @ beta``."""
    calc = Calc(lambda x, beta: x @ beta, x=x, beta=beta)
    var = Var(calc, distribution, name)
    var.role = "Smooth"
    return var
